# encoding: utf-8
"""
@author: BrikerMan
@contact: eliyar917@gmail.com
@blog: https://eliyar.biz

@version: 1.0
@license: Apache Licence
@file: crf_accuracies.py.py
@time: 2019-01-23 17:10

"""
from keras import backend as K

"""
from https://github.com/keras-team/keras-contrib
"""


def _get_accuracy(y_true, y_pred, mask, sparse_target=False):
    y_pred = K.argmax(y_pred, -1)
    if sparse_target:
        y_true = K.cast(y_true[:, :, 0], K.dtype(y_pred))
    else:
        y_true = K.argmax(y_true, -1)
    judge = K.cast(K.equal(y_pred, y_true), K.floatx())
    if mask is None:
        return K.mean(judge)
    else:
        mask = K.cast(mask, K.floatx())
        return K.sum(judge * mask) / K.sum(mask)


def crf_viterbi_accuracy(y_true, y_pred):
    '''Use Viterbi algorithm to get best path, and compute its accuracy.
    `y_pred` must be an output from CRF.'''
    crf, idx = y_pred._keras_history[:2]
    X = crf._inbound_nodes[idx].input_tensors[0]
    mask = crf._inbound_nodes[idx].input_masks[0]
    y_pred = crf.viterbi_decoding(X, mask)
    return _get_accuracy(y_true, y_pred, mask, crf.sparse_target)


def crf_marginal_accuracy(y_true, y_pred):
    '''Use time-wise marginal argmax as prediction.
    `y_pred` must be an output from CRF with `learn_mode="marginal"`.'''
    crf, idx = y_pred._keras_history[:2]
    X = crf._inbound_nodes[idx].input_tensors[0]
    mask = crf._inbound_nodes[idx].input_masks[0]
    y_pred = crf.get_marginal_prob(X, mask)
    return _get_accuracy(y_true, y_pred, mask, crf.sparse_target)


def crf_accuracy(y_true, y_pred):
    '''Ge default accuracy based on CRF `test_mode`.'''
    crf, idx = y_pred._keras_history[:2]
    if crf.test_mode == 'viterbi':
        return crf_viterbi_accuracy(y_true, y_pred)
    else:
        return crf_marginal_accuracy(y_true, y_pred)


if __name__ == "__main__":
    print("Hello world")



# encoding: utf-8
"""
@author: BrikerMan
@contact: eliyar917@gmail.com
@blog: https://eliyar.biz

@version: 1.0
@license: Apache Licence
@file: logger.py
@time: 2019-01-19 09:58

"""
import os
import sys
import logging
from colorlog import ColoredFormatter


def init_logger():
    level = os.getenv('LOG_LEVEL', 'DEBUG')
    change_log_level(level)


def change_log_level(log_level):
    print('----------------------')
    # color_format = "%(log_color)s[%(asctime)s] %(levelname)-7s  " \
    #                "%(name)s:%(filename)s:%(lineno)d - %(message)s"
    color_format = "%(log_color)s[%(asctime)s] %(levelname)-5s " \
                   "- %(message)s"

    color_formatter = ColoredFormatter(color_format,
                                       datefmt=None,
                                       reset=True,
                                       log_colors={
                                           'DEBUG': 'white',
                                           'INFO': 'green',
                                           'WARNING': 'purple',
                                           'ERROR': 'red',
                                           'CRITICAL': 'red,bg_white',
                                       },
                                       secondary_log_colors={},
                                       style='%'
                                       )

    print_handler = logging.StreamHandler(sys.stdout)
    print_handler.setFormatter(color_formatter)
    print_handler.setLevel(log_level)

    logging.basicConfig(level=logging.DEBUG,
                        handlers=[
                            # handler,
                            print_handler
                        ])

    logging.info('logging init finished')


if __name__ == "__main__":
    init_logger()

    logging.info('info')
    logging.error('error')
    logging.warning('warning')




# encoding: utf-8
"""
@author: BrikerMan
@contact: eliyar917@gmail.com
@blog: https://eliyar.biz

@version: 1.0
@license: Apache Licence
@file: helper.py
@time: 2019-01-19 16:25

"""
import logging
import os
import random
from typing import List, Optional

import download
import h5py
import numpy as np
from keras import backend as K
from keras.layers import Layer
from keras.preprocessing import sequence
from keras.utils import to_categorical

from kashgari.macros import DATA_PATH
from kashgari.macros import STORAGE_HOST


# def h5f_generator(h5path: str,
#                   # indices: List[int],
#                   num_classes: int,
#                   batch_size: int = 128):
#     """
#     fit generator for h5 file
#     :param h5path: target f5file
#     :param num_classes: label counts to covert y label to one hot array
#     :param batch_size:
#     :return:
#     """
#
#     db = h5py.File(h5path, "r")
#     while True:
#         page_list = list(range(len(db['x']) // batch_size + 1))
#         random.shuffle(page_list)
#         for page in page_list:
#             x = db["x"][page: (page + 1) * batch_size]
#             y = to_categorical(db["y"][page: (page + 1) * batch_size],
#                                num_classes=num_classes,
#                                dtype=np.int)
#             yield (x, y)


def unison_shuffled_copies(a, b):
    assert len(a) == len(b)
    c = list(zip(a, b))
    random.shuffle(c)
    a, b = zip(*c)
    return list(a), list(b)


def weighted_categorical_crossentropy(weights):
    """
    A weighted version of keras.objectives.categorical_crossentropy

    Variables:
        weights: numpy array of shape (C,) where C is the number of classes

    Usage:
        weights = np.array([0.5,2,10]) # Class one at 0.5, class 2 twice the normal weights, class 3 10x.
        loss = weighted_categorical_crossentropy(weights)
        model.compile(loss=loss,optimizer='adam')
    """

    weights = K.variable(weights)

    def loss(y_true, y_pred):
        # scale predictions so that the class probas of each sample sum to 1
        y_pred /= K.sum(y_pred, axis=-1, keepdims=True)
        # clip to prevent NaN's and Inf's
        y_pred = K.clip(y_pred, K.epsilon(), 1 - K.epsilon())
        # calc
        loss = y_true * K.log(y_pred) * weights
        loss = -K.sum(loss, -1)
        return loss

    return loss


def cached_path(file_path: str, download_url: Optional[str], sub_folders: List[str] = []):
    if os.path.exists(file_path):
        return file_path

    file_name_list = [DATA_PATH] + sub_folders + [file_path]
    file_path = os.path.join(*file_name_list)
    if os.path.exists(file_path):
        return file_path

    if download_url.startswith('http'):
        url = download_url
    else:
        url = STORAGE_HOST + download_url

    final_path = file_path
    if url.endswith('zip'):
        kind = 'zip'
        download_path = os.path.dirname(file_path)
    elif url.endswith('tar.gz'):
        kind = 'tar.gz'
        download_path = os.path.dirname(file_path)
    else:
        kind = 'file'
        download_path = file_path
    # url = url.replace('https://', 'http://')
    logging.info('start downloading file, if it takes too long, you could download with other downloader\n'
                 'url  : {}\n'
                 'path : {}'.format(url, file_path))
    e_path = download.download(url, download_path, kind=kind, replace=True)
    logging.info('downloader file_path {}, {} '.format(e_path, file_path))
    return final_path
    # if file_path.endswith('.bz2'):
    #     archive_path = e_path
    #     outfile_path = e_path[:-4]
    #     with open(archive_path, 'rb') as source, open(outfile_path, 'wb') as dest:
    #         dest.write(bz2.decompress(source.read()))
    #     return outfile_path
    # else:
    #     return final_path


# def check_should_download(file: str,
#                           download_url: Optional[str],
#                           sub_folders: List[str] = None):
#     """
#     check should download the file, if exist return file url, if not download and unzip
#     :param file:
#     :param sub_folders:
#     :param download_url:
#     :return:
#     """
#     logging.debug('check_should_download: file {}\ndownload_url {}\nsub_folders {}'.format(file,
#                                                                                            download_url,
#                                                                                            sub_folders))
#     if sub_folders is None:
#         sub_folders = []
#
#     if os.path.exists(file):
#         return file
#
#     folders = [DATA_PATH] + sub_folders + [file]
#     target_path = os.path.join(*folders)
#     original_file_path = target_path
#
#     if os.path.exists(target_path):
#         return target_path
#
#     if not download_url:
#         raise ValueError("need to provide valid model name or path")
#
#     if download_url.startswith('http'):
#         url = download_url
#     else:
#         url = STORAGE_HOST + download_url
#
#     if url.endswith('zip'):
#         kind = 'zip'
#     elif url.endswith('tar.gz'):
#         kind = 'tar.gz'
#     else:
#         kind = 'file'
#         target_path = os.path.join(target_path, url.split('/')[-1])
#
#     logging.info('start downloading file, if it takes too long, you could download with other downloader\n'
#                  'url  : {}\n'
#                  'path : {}'.format(url,
#                                     os.path.dirname(target_path)))
#
#     file_path = download.download(url, target_path, kind=kind, replace=True)
#     logging.debug('file downloaded to {}'.format(file_path))
#     if file_path.endswith('.bz2'):
#         archive_path = file_path
#         outfile_path = file_path[:-4]
#         with open(archive_path, 'rb') as source, open(outfile_path, 'wb') as dest:
#             dest.write(bz2.decompress(source.read()))
#         return original_file_path
#     else:
#         return target_path


if __name__ == "__main__":
    from kashgari.utils.logger import init_logger
    init_logger()
    file = 'embedding/word2vec/sgns.weibo.bigram-char'
    url = 'embedding/word2vec/sgns.weibo.bigram-char.bz2'
    print(cached_path(file, url))



# encoding: utf-8
"""
@author: BrikerMan
@contact: eliyar917@gmail.com
@blog: https://eliyar.biz

@version: 1.0
@license: Apache Licence
@file: crf_losss.py
@time: 2019-01-23 17:08

"""
from keras import backend as K
from keras.losses import categorical_crossentropy
from keras.losses import sparse_categorical_crossentropy

"""
from https://github.com/keras-team/keras-contrib
"""


def crf_nll(y_true, y_pred):
    """The negative log-likelihood for linear chain Conditional Random Field (CRF).

    This loss function is only used when the `layers.CRF` layer
    is trained in the "join" mode.

    # Arguments
        y_true: tensor with true targets.
        y_pred: tensor with predicted targets.

    # Returns
        A scalar representing corresponding to the negative log-likelihood.

    # Raises
        TypeError: If CRF is not the last layer.

    # About GitHub
        If you open an issue or a pull request about CRF, please
        add `cc @lzfelix` to notify Luiz Felix.
    """

    crf, idx = y_pred._keras_history[:2]
    if crf._outbound_nodes:
        raise TypeError('When learn_model="join", CRF must be the last layer.')
    if crf.sparse_target:
        y_true = K.one_hot(K.cast(y_true[:, :, 0], 'int32'), crf.units)
    X = crf._inbound_nodes[idx].input_tensors[0]
    mask = crf._inbound_nodes[idx].input_masks[0]
    nloglik = crf.get_negative_log_likelihood(y_true, X, mask)
    return nloglik


def crf_loss(y_true, y_pred):
    """General CRF loss function depending on the learning mode.

    # Arguments
        y_true: tensor with true targets.
        y_pred: tensor with predicted targets.

    # Returns
        If the CRF layer is being trained in the join mode, returns the negative
        log-likelihood. Otherwise returns the categorical crossentropy implemented
        by the underlying Keras backend.

    # About GitHub
        If you open an issue or a pull request about CRF, please
        add `cc @lzfelix` to notify Luiz Felix.
    """
    crf, idx = y_pred._keras_history[:2]
    if crf.learn_mode == 'join':
        return crf_nll(y_true, y_pred)
    else:
        if crf.sparse_target:
            return sparse_categorical_crossentropy(y_true, y_pred)
        else:
            return categorical_crossentropy(y_true, y_pred)


# encoding: utf-8
"""
@author: BrikerMan
@contact: eliyar917@gmail.com
@blog: https://eliyar.biz

@version: 1.0
@license: Apache Licence
@file: base_model.py
@time: 2019-01-19 11:50

"""
import logging
import random
from typing import Tuple, Dict

import numpy as np
from keras.preprocessing import sequence
from keras.utils import to_categorical
from sklearn import metrics
from sklearn.utils import class_weight as class_weight_calculte
from sklearn.preprocessing import MultiLabelBinarizer

from kashgari import macros as k
from kashgari.tasks.base import BaseModel
from kashgari.embeddings import BaseEmbedding
from kashgari.type_hints import *


class ClassificationModel(BaseModel):

    def __init__(self,
                 embedding: BaseEmbedding = None,
                 hyper_parameters: Dict = None,
                 multi_label: bool = False,
                 **kwargs):
        """

        :param embedding:
        :param hyper_parameters:
        :param multi_label:
        :param kwargs:
        """
        super(ClassificationModel, self).__init__(embedding, hyper_parameters, **kwargs)
        self.multi_label = multi_label
        self.multi_label_binarizer: MultiLabelBinarizer = None

        if self.multi_label:
            if not hyper_parameters or \
                    hyper_parameters.get('compile_params', {}).get('loss') is None:
                self.hyper_parameters['compile_params']['loss'] = 'binary_crossentropy'
            else:
                logging.warning('recommend to use binary_crossentropy loss for multi_label task')

            if not hyper_parameters or \
                    hyper_parameters.get('compile_params', {}).get('metrics') is None:
                self.hyper_parameters['compile_params']['metrics'] = ['categorical_accuracy']
            else:
                logging.warning('recommend to use categorical_accuracy metrivs for multi_label task')

            if not hyper_parameters or \
                    hyper_parameters.get('activation_layer', {}).get('sigmoid') is None:
                self.hyper_parameters['activation_layer']['activation'] = 'sigmoid'
            else:
                logging.warning('recommend to use sigmoid activation for multi_label task')

    def info(self):
        info = super(ClassificationModel, self).info()
        info['model_info']['multi_label'] = self.multi_label
        return info

    @property
    def label2idx(self) -> Dict[str, int]:
        return self._label2idx

    @property
    def token2idx(self) -> Dict[str, int]:
        return self.embedding.token2idx

    @label2idx.setter
    def label2idx(self, value):
        self._label2idx = value
        self._idx2label = dict([(val, key) for (key, val) in value.items()])

    def build_model(self):
        """
        build model function
        :return:
        """
        raise NotImplementedError()

    @classmethod
    def load_model(cls, model_path: str):
        agent: ClassificationModel = super(ClassificationModel, cls).load_model(model_path)
        agent.multi_label = agent.model_info.get('multi_label', False)
        if agent.multi_label:
            keys = list(agent.label2idx.keys())
            agent.multi_label_binarizer = MultiLabelBinarizer(classes=keys)
            agent.multi_label_binarizer.fit(keys[0])
        return agent

    def build_token2id_label2id_dict(self,
                                     x_train: List[List[str]],
                                     y_train: List[str],
                                     x_validate: List[List[str]] = None,
                                     y_validate: List[str] = None):
        if x_validate:
            x_data = x_train + x_validate
            y_data = y_train + y_validate
        else:
            x_data = x_train
            y_data = y_train
        self.embedding.build_token2idx_dict(x_data, 3)

        if self.multi_label:
            label_set = set()
            for i in y_data:
                label_set = label_set.union(list(i))
        else:
            label_set = set(y_data)

        label2idx = {}
        for idx, label in enumerate(label_set):
            label2idx[label] = idx
        self._label2idx = label2idx
        self._idx2label = dict([(val, key) for (key, val) in label2idx.items()])
        self.multi_label_binarizer = MultiLabelBinarizer(classes=list(self.label2idx.keys()))

    def convert_label_to_idx(self, label: Union[List[str], str]) -> Union[List[int], int]:
        if isinstance(label, str):
            return self.label2idx[label]
        else:
            return [self.label2idx[l] for l in label]

    def convert_idx_to_label(self, token: Union[List[int], int]) -> Union[List[str], str]:
        if isinstance(token, int):
            return self._idx2label[token]
        else:
            return [self._idx2label[l] for l in token]

    def get_data_generator(self,
                           x_data: List[List[str]],
                           y_data: List[str],
                           batch_size: int = 64,
                           is_bert: bool = False):
        while True:
            page_list = list(range((len(x_data) // batch_size) + 1))
            random.shuffle(page_list)
            for page in page_list:
                start_index = page * batch_size
                end_index = start_index + batch_size
                target_x = x_data[start_index: end_index]
                target_y = y_data[start_index: end_index]
                if len(target_x) == 0:
                    target_x = x_data[0: batch_size]
                    target_y = y_data[0: batch_size]

                tokenized_x = self.embedding.tokenize(target_x)

                padded_x = sequence.pad_sequences(tokenized_x,
                                                  maxlen=self.embedding.sequence_length,
                                                  padding='post')

                if self.multi_label:
                    padded_y = self.multi_label_binarizer.fit_transform(target_y)
                else:
                    tokenized_y = self.convert_label_to_idx(target_y)
                    padded_y = to_categorical(tokenized_y,
                                              num_classes=len(self.label2idx),
                                              dtype=np.int)
                if is_bert:
                    padded_x_seg = np.zeros(shape=(len(padded_x), self.embedding.sequence_length))
                    x_input_data = [padded_x, padded_x_seg]
                else:
                    x_input_data = padded_x
                yield (x_input_data, padded_y)

    def fit(self,
            x_train: List[List[str]],
            y_train: Union[List[str], List[List[str]], List[Tuple[str]]],
            x_validate: List[List[str]] = None,
            y_validate: Union[List[str], List[List[str]], List[Tuple[str]]] = None,
            batch_size: int = 64,
            epochs: int = 5,
            class_weight: bool = False,
            fit_kwargs: Dict = None,
            **kwargs):
        """

        :param x_train: list of training data.
        :param y_train: list of training target label data.
        :param x_validate: list of validation data.
        :param y_validate: list of validation target label data.
        :param batch_size: batch size for trainer model
        :param epochs: Number of epochs to train the model.
        :param class_weight: set class weights for imbalanced classes
        :param fit_kwargs: additional kwargs to be passed to
               :func:`~keras.models.Model.fit`
        :param kwargs:
        :return:
        """
        assert len(x_train) == len(y_train)
        self.build_token2id_label2id_dict(x_train, y_train, x_validate, y_validate)

        if len(x_train) < batch_size:
            batch_size = len(x_train) // 2

        if not self.model:
            if self.embedding.sequence_length == 0:
                self.embedding.sequence_length = sorted([len(x) for x in x_train])[int(0.95 * len(x_train))]
                logging.info('sequence length set to {}'.format(self.embedding.sequence_length))
            self.build_model()

        train_generator = self.get_data_generator(x_train,
                                                  y_train,
                                                  batch_size,
                                                  is_bert=self.embedding.is_bert)

        if fit_kwargs is None:
            fit_kwargs = {}

        if x_validate:
            validation_generator = self.get_data_generator(x_validate,
                                                           y_validate,
                                                           batch_size,
                                                           is_bert=self.embedding.is_bert)
            fit_kwargs['validation_data'] = validation_generator
            fit_kwargs['validation_steps'] = max(len(x_validate) // batch_size, 1)

        if class_weight:
            y_list = self.convert_label_to_idx(y_train)
            class_weights = class_weight_calculte.compute_class_weight('balanced',
                                                                       np.unique(y_list),
                                                                       y_list)
        else:
            class_weights = None

        self.model.fit_generator(train_generator,
                                 steps_per_epoch=len(x_train) // batch_size,
                                 epochs=epochs,
                                 class_weight=class_weights,
                                 **fit_kwargs)

    def _format_output_dic(self, words: List[str], res: np.ndarray):
        results = sorted(list(enumerate(res)), key=lambda x: -x[1])
        candidates = []
        for result in results:
            candidates.append({
                'name': self.convert_idx_to_label([result[0]])[0],
                'confidence': float(result[1]),
            })

        data = {
            'words': words,
            'class': candidates[0],
            'class_candidates': candidates
        }
        return data

    def predict(self,
                sentence: Union[List[str], List[List[str]]],
                batch_size=None,
                output_dict=False,
                multi_label_threshold=0.6,
                debug_info=False) -> Union[List[str], str, List[Dict], Dict]:
        """
        predict with model
        :param sentence: single sentence as List[str] or list of sentence as List[List[str]]
        :param batch_size: predict batch_size
        :param output_dict: return dict with result with confidence
        :param multi_label_threshold:
        :param debug_info: print debug info using logging.debug when True
        :return:
        """
        tokens = self.embedding.tokenize(sentence)
        is_list = not isinstance(sentence[0], str)
        if is_list:
            padded_tokens = sequence.pad_sequences(tokens,
                                                   maxlen=self.embedding.sequence_length,
                                                   padding='post')
        else:
            padded_tokens = sequence.pad_sequences([tokens],
                                                   maxlen=self.embedding.sequence_length,
                                                   padding='post')
        if self.embedding.is_bert:
            x = [padded_tokens, np.zeros(shape=(len(padded_tokens), self.embedding.sequence_length))]
        else:
            x = padded_tokens
        res = self.model.predict(x, batch_size=batch_size)

        if self.multi_label:
            if debug_info:
                logging.info('raw output: {}'.format(res))
            res[res >= multi_label_threshold] = 1
            res[res < multi_label_threshold] = 0
            predict_result = res
        else:
            predict_result = res.argmax(-1)

        if debug_info:
            logging.info('input: {}'.format(x))
            logging.info('output: {}'.format(res))
            logging.info('output argmax: {}'.format(predict_result))

        if output_dict:
            if is_list:
                words_list: List[List[str]] = sentence
            else:
                words_list: List[List[str]] = [sentence]
            results = []
            for index in range(len(words_list)):
                results.append(self._format_output_dic(words_list[index], res[index]))
            if is_list:
                return results
            else:
                return results[0]
        else:
            if self.multi_label:
                results = self.multi_label_binarizer.inverse_transform(predict_result)
            else:
                results = self.convert_idx_to_label(predict_result)
            if is_list:
                return results
            else:
                return results[0]

    def evaluate(self, x_data, y_data, batch_size=None, digits=4, debug_info=False) -> Tuple[float, float, Dict]:
        y_pred = self.predict(x_data, batch_size=batch_size)
        report = metrics.classification_report(y_data, y_pred, output_dict=True, digits=digits)
        print(metrics.classification_report(y_data, y_pred, digits=digits))
        if debug_info:
            for index in random.sample(list(range(len(x_data))), 5):
                logging.debug('------ sample {} ------'.format(index))
                logging.debug('x      : {}'.format(x_data[index]))
                logging.debug('y      : {}'.format(y_data[index]))
                logging.debug('y_pred : {}'.format(y_pred[index]))
        return report


# encoding: utf-8
"""
@author: BrikerMan
@contact: eliyar917@gmail.com
@blog: https://eliyar.biz

@version: 1.0
@license: Apache Licence
@file: base_model
@time: 2019-01-21

"""
import random
import logging
from typing import Tuple, Dict

import numpy as np
from keras.preprocessing import sequence
from keras.utils import to_categorical
from seqeval.metrics import classification_report
from seqeval.metrics.sequence_labeling import get_entities

import kashgari.macros as k
from kashgari.utils import helper
from kashgari.type_hints import *

from kashgari.tasks.base import BaseModel
from kashgari.embeddings import BaseEmbedding


class SequenceLabelingModel(BaseModel):

    def __init__(self, embedding: BaseEmbedding = None, hyper_parameters: Dict = None, **kwargs):
        super(SequenceLabelingModel, self).__init__(embedding, hyper_parameters, **kwargs)
        self.task = 'sequence_labeling'

    @property
    def label2idx(self) -> Dict[str, int]:
        return self._label2idx

    @property
    def token2idx(self) -> Dict[str, int]:
        return self.embedding.token2idx

    @label2idx.setter
    def label2idx(self, value):
        self._label2idx = value
        self._idx2label = dict([(val, key) for (key, val) in value.items()])

    def build_model(self, loss_f=None, optimizer=None, metrics=None, **kwargs):
        """
        build model function
        :return:
        """
        raise NotImplementedError()

    def build_token2id_label2id_dict(self,
                                     x_train: List[List[str]],
                                     y_train: List[List[str]],
                                     x_validate: List[List[str]] = None,
                                     y_validate: List[List[str]] = None):
        for index in range(len(x_train)):
            assert len(x_train[index]) == len(y_train[index])
        x_data = x_train
        y_data = y_train
        if x_validate:
            x_data = x_train + x_validate
            y_data = y_data + y_validate
        self.embedding.build_token2idx_dict(x_data, 3)

        label_set = []
        for seq in y_data:
            for y in seq:
                if y not in label_set:
                    label_set.append(y)

        label2idx = {
            k.PAD: 0,
            k.BOS: 1,
            k.EOS: 2
        }
        label_set = [i for i in label_set if i not in label2idx]
        for label in label_set:
            label2idx[label] = len(label2idx)

        self.label2idx = label2idx

    def convert_labels_to_idx(self,
                              label: Union[List[List[str]], List[str]],
                              add_eos_bos: bool = True) -> Union[List[List[int]], List[int]]:

        def tokenize_tokens(seq: List[str]):
            tokens = [self._label2idx[i] for i in seq]
            if add_eos_bos:
                tokens = [self._label2idx[k.BOS]] + tokens + [self._label2idx[k.EOS]]
            return tokens

        if isinstance(label[0], str):
            return tokenize_tokens(label)
        else:
            return [tokenize_tokens(l) for l in label]

    def convert_idx_to_labels(self,
                              idx: Union[List[List[int]], List[int]],
                              tokens_length: Union[List[int], int],
                              remove_eos_bos: bool = True) -> Union[List[str], str]:

        def reverse_tokenize_tokens(idx_item, seq_length):
            if remove_eos_bos:
                seq = idx_item[1: 1 + seq_length]
            else:
                seq = idx_item
            tokens = [self._idx2label[i] for i in seq]
            return tokens

        if isinstance(idx[0], int):
            return reverse_tokenize_tokens(idx, tokens_length)
        else:
            labels = []
            for index in range(len(idx)):
                idx_item = idx[index]
                seq_length = tokens_length[index]
                labels.append(reverse_tokenize_tokens(idx_item, seq_length))
            return labels

    def get_data_generator(self,
                           x_data: List[List[str]],
                           y_data: List[List[str]],
                           batch_size: int = 64):
        is_bert = self.embedding.embedding_type == 'bert'
        while True:
            page_list = list(range((len(x_data) // batch_size) + 1))
            random.shuffle(page_list)
            for page in page_list:
                start_index = page * batch_size
                end_index = start_index + batch_size
                target_x = x_data[start_index: end_index]
                target_y = y_data[start_index: end_index]
                if len(target_x) == 0:
                    target_x = x_data[0: batch_size]
                    target_y = y_data[0: batch_size]

                tokenized_x = self.embedding.tokenize(target_x)
                tokenized_y = self.convert_labels_to_idx(target_y)

                padded_x = sequence.pad_sequences(tokenized_x,
                                                  maxlen=self.embedding.sequence_length,
                                                  padding='post', truncating='post')
                padded_y = sequence.pad_sequences(tokenized_y,
                                                  maxlen=self.embedding.sequence_length,
                                                  padding='post', truncating='post')

                one_hot_y = to_categorical(padded_y, num_classes=len(self.label2idx))

                if is_bert:
                    padded_x_seg = np.zeros(shape=(len(padded_x), self.embedding.sequence_length))
                    x_input_data = [padded_x, padded_x_seg]
                else:
                    x_input_data = padded_x
                yield (x_input_data, one_hot_y)

    def fit(self,
            x_train: List[List[str]],
            y_train: List[List[str]],
            x_validate: List[List[str]] = None,
            y_validate: List[List[str]] = None,
            batch_size: int = 64,
            epochs: int = 5,
            labels_weight: bool = None,
            default_labels_weight: float = 50.0,
            fit_kwargs: Dict = None,
            **kwargs):
        """

        :param x_train: list of training data.
        :param y_train: list of training target label data.
        :param batch_size: batch size for trainer model
        :param epochs: Number of epochs to train the model.
        :param x_validate: list of validation data.
        :param y_validate: list of validation target label data.
        :param y_validate: list of validation target label data.
        :param y_validate: list of validation target label data.
        :param labels_weight: set class weights for imbalanced classes
        :param default_labels_weight: default weight for labels not in labels_weight dict
        :param fit_kwargs: additional kwargs to be passed to
               :func:`~keras.models.Model.fit`
        :return:
        """
        assert len(x_train) == len(y_train)
        self.build_token2id_label2id_dict(x_train, y_train, x_validate, y_validate)

        if len(x_train) < batch_size:
            batch_size = len(x_train) // 2

        if not self.model:
            if self.embedding.sequence_length == 0:
                self.embedding.sequence_length = sorted([len(x) for x in x_train])[int(0.95 * len(x_train))]
                logging.info('sequence length set to {}'.format(self.embedding.sequence_length))

            if labels_weight:
                weights = []
                initial_weights = {
                    k.PAD: 1,
                    k.BOS: 1,
                    k.EOS: 1,
                    'O': 1
                }
                for label in self.label2idx.keys():
                    weights.append(initial_weights.get(label, default_labels_weight))
                loss_f = helper.weighted_categorical_crossentropy(np.array(weights))
                self.model_info['loss'] = {
                    'func': 'weighted_categorical_crossentropy',
                    'weights': weights
                }

                self.build_model(loss_f=loss_f, metrics=['categorical_accuracy', 'acc'])
            else:
                self.build_model()

        train_generator = self.get_data_generator(x_train,
                                                  y_train,
                                                  batch_size)

        if fit_kwargs is None:
            fit_kwargs = {}

        if x_validate:
            validation_generator = self.get_data_generator(x_validate,
                                                           y_validate,
                                                           batch_size)

            fit_kwargs['validation_data'] = validation_generator
            fit_kwargs['validation_steps'] = len(x_validate) // batch_size

        self.model.fit_generator(train_generator,
                                 steps_per_epoch=len(x_train) // batch_size,
                                 epochs=epochs,
                                 **fit_kwargs)

    def _format_output_dic(self, words: List[str], tags: List[str], chunk_joiner: str):
        chunks = get_entities(tags)
        res = {
            'words': words,
            'entities': []
        }
        for chunk_type, chunk_start, chunk_end in chunks:
            chunk_end += 1
            entity = {
                'text': chunk_joiner.join(words[chunk_start: chunk_end]),
                'type': chunk_type,
                # 'score': float(np.average(prob[chunk_start: chunk_end])),
                'beginOffset': chunk_start,
                'endOffset': chunk_end
            }
            res['entities'].append(entity)
        return res

    def predict(self,
                sentence: Union[List[str], List[List[str]]],
                batch_size=None,
                output_dict=False,
                chunk_joiner=' ',
                debug_info=False):
        """
        predict with model
        :param sentence: input for predict, accept a single sentence as type List[str] or
                         list of sentence as List[List[str]]
        :param batch_size: predict batch_size
        :param output_dict: return dict with result with confidence
        :param chunk_joiner: the char to join the chunks when output dict
        :param debug_info: print debug info using logging.debug when True
        :return:
        """
        tokens = self.embedding.tokenize(sentence)
        is_list = not isinstance(sentence[0], str)
        if is_list:
            seq_length = [len(item) for item in sentence]
            padded_tokens = sequence.pad_sequences(tokens,
                                                   maxlen=self.embedding.sequence_length,
                                                   padding='post', truncating='post')
        else:
            seq_length = [len(sentence)]
            padded_tokens = sequence.pad_sequences([tokens],
                                                   maxlen=self.embedding.sequence_length,
                                                   padding='post', truncating='post')
        if self.embedding.is_bert:
            x = [padded_tokens, np.zeros(shape=(len(padded_tokens), self.embedding.sequence_length))]
        else:
            x = padded_tokens

        predict_result_prob = self.model.predict(x, batch_size=batch_size)
        predict_result = predict_result_prob.argmax(-1)
        if debug_info:
            logging.info('input: {}'.format(x))
            logging.info('output: {}'.format(predict_result_prob))
            logging.info('output argmax: {}'.format(predict_result))

        result: List[List[str]] = self.convert_idx_to_labels(predict_result, seq_length)
        if output_dict:
            dict_list = []
            if is_list:
                sentence_list: List[List[str]] = sentence
            else:
                sentence_list: List[List[str]] = [sentence]
            for index in range(len(sentence_list)):
                dict_list.append(self._format_output_dic(sentence_list[index],
                                                         result[index],
                                                         chunk_joiner))
            if is_list:
                return dict_list
            else:
                return dict_list[0]
        else:
            if is_list:
                return result
            else:
                return result[0]

    def evaluate(self, x_data, y_data, batch_size=None, digits=4, debug_info=False) -> Tuple[float, float, Dict]:
        seq_length = [len(x) for x in x_data]
        tokenized_y = self.convert_labels_to_idx(y_data)
        padded_y = sequence.pad_sequences(tokenized_y,
                                          maxlen=self.embedding.sequence_length,
                                          padding='post', truncating='post')
        y_true = self.convert_idx_to_labels(padded_y, seq_length)
        y_pred = self.predict(x_data, batch_size=batch_size)
        if debug_info:
            for index in random.sample(list(range(len(x_data))), 5):
                logging.debug('------ sample {} ------'.format(index))
                logging.debug('x      : {}'.format(x_data[index]))
                logging.debug('y_true : {}'.format(y_true[index]))
                logging.debug('y_pred : {}'.format(y_pred[index]))
        report = classification_report(y_true, y_pred, digits=digits)
        print(classification_report(y_true, y_pred, digits=digits))
        return report



# encoding: utf-8
"""
@author: BrikerMan
@contact: eliyar917@gmail.com
@blog: https://eliyar.biz

@version: 1.0
@license: Apache Licence
@file: macros.py
@time: 2019-01-19 09:58

"""
import bz2
import os
import pathlib
from enum import Enum
from pathlib import Path

import download

PAD = "[PAD]"
BOS = "[BOS]"
EOS = "[EOS]"
UNK = "[UNK]"

MARKED_KEYS = [PAD, BOS, EOS, UNK]

NO_TAG = 'O'

home = str(Path.home())

DATA_PATH = os.path.join(home, '.kashgari')
STORAGE_HOST = 'http://storage.eliyar.biz/'
PROCESSED_CORPUS_PATH = os.path.join(DATA_PATH, 'pre_processed')

pathlib.Path(PROCESSED_CORPUS_PATH).mkdir(parents=True, exist_ok=True)


class _Config(object):
    def __init__(self):
        self.use_CuDNN_cell = False


config = _Config()


class CustomEmbedding(object):
    def __init__(self, embedding_size=100):
        self.embedding_size = embedding_size


class TaskType(Enum):
    classification = 'classification'
    tagging = 'tagging'


class DataSetType(Enum):
    train = 'train'
    test = 'test'
    validate = 'validate'


class SegmenterType(Enum):
    space = 'space'
    jieba = 'jieba'
    char = 'char'


URL_MAP = {
    'w2v.sgns.weibo.bigram': 'embedding/word2vev/sgns.weibo.bigram.bz2'
}


def download_file(file: str):
    url = STORAGE_HOST + file
    target_path = os.path.join(DATA_PATH, file)
    download.download(url, target_path)


def download_if_not_existed(file_path: str) -> str:
    target_path = os.path.join(DATA_PATH, file_path)
    if not os.path.exists(target_path[:-4]):
        download_file(file_path)
        with open(target_path, 'rb') as source, open(target_path[:-4], 'wb') as dest:
            dest.write(bz2.decompress(source.read()))
    return target_path[:-4]


def get_model_path(file: str) -> str:
    file_path = URL_MAP.get(file, file)
    return download_if_not_existed(file_path)


if __name__ == "__main__":
    from kashgari.utils.logger import init_logger
    init_logger()



# encoding: utf-8
"""
@author: Alex
@contact: ialexwwang@gmail.com

@version: 0.1
@license: Apache Licence
@file: deep_models.py
@time: 2019-02-21 17:54

@Reference: https://github.com/zake7749/DeepToxic/blob/master/sotoxic/models/keras/model_zoo.py
"""
from __future__ import absolute_import, division

import keras
from keras.layers import Bidirectional, Conv1D
from keras.layers import Dense, Lambda, Flatten
from keras.layers import Dropout, SpatialDropout1D
from keras.layers import GlobalAveragePooling1D, GlobalMaxPooling1D, MaxPooling1D
from keras.layers import concatenate
from keras.models import Model

from kashgari.layers import AttentionWeightedAverage, KMaxPooling, LSTMLayer, GRULayer
from kashgari.tasks.classification.base_model import ClassificationModel


class CNNModel(ClassificationModel):
    __architect_name__ = 'CNNModel'
    __base_hyper_parameters__ = {
        'conv1d_layer': {
            'filters': 128,
            'kernel_size': 5,
            'activation': 'relu'
        },
        'max_pool_layer': {},
        'dense_1_layer': {
            'units': 64,
            'activation': 'relu'
        },
        'activation_layer': {
            'activation': 'softmax'
        },
        'optimizer': {
            'module': 'keras.optimizers',
            'name': 'Adam',
            'params': {
                'lr': 1e-3,
                'decay': 0.0
            }
        },
        'compile_params': {
            'loss': 'categorical_crossentropy',
            # 'optimizer': 'adam',
            'metrics': ['accuracy']
        }
    }

    def build_model(self):
        base_model = self.embedding.model
        conv1d_layer = Conv1D(**self.hyper_parameters['conv1d_layer'])(base_model.output)
        max_pool_layer = GlobalMaxPooling1D(**self.hyper_parameters['max_pool_layer'])(conv1d_layer)
        dense_1_layer = Dense(**self.hyper_parameters['dense_1_layer'])(max_pool_layer)
        dense_2_layer = Dense(len(self.label2idx), **self.hyper_parameters['activation_layer'])(dense_1_layer)

        model = Model(base_model.inputs, dense_2_layer)
        optimizer = getattr(eval(self.hyper_parameters['optimizer']['module']),
                            self.hyper_parameters['optimizer']['name'])(
            **self.hyper_parameters['optimizer']['params'])
        model.compile(optimizer=optimizer, **self.hyper_parameters['compile_params'])
        self.model = model
        self.model.summary()


class BLSTMModel(ClassificationModel):
    __architect_name__ = 'BLSTMModel'
    __base_hyper_parameters__ = {
        'lstm_layer': {
            'units': 256,
            'return_sequences': False
        },
        'activation_layer': {
            'activation': 'softmax'
        },
        'optimizer': {
            'module': 'keras.optimizers',
            'name': 'Adam',
            'params': {
                'lr': 1e-3,
                'decay': 0.0
            }
        },
        'compile_params': {
            'loss': 'categorical_crossentropy',
            # 'optimizer': 'adam',
            'metrics': ['accuracy']
        }
    }

    def build_model(self):
        base_model = self.embedding.model
        blstm_layer = Bidirectional(LSTMLayer(**self.hyper_parameters['lstm_layer']))(base_model.output)
        dense_layer = Dense(len(self.label2idx), **self.hyper_parameters['activation_layer'])(blstm_layer)
        output_layers = [dense_layer]

        model = Model(base_model.inputs, output_layers)
        optimizer = getattr(eval(self.hyper_parameters['optimizer']['module']),
                            self.hyper_parameters['optimizer']['name'])(
            **self.hyper_parameters['optimizer']['params'])
        model.compile(optimizer=optimizer, **self.hyper_parameters['compile_params'])
        self.model = model
        self.model.summary()


class CNNLSTMModel(ClassificationModel):
    __architect_name__ = 'CNNLSTMModel'
    __base_hyper_parameters__ = {
        'conv_layer': {
            'filters': 32,
            'kernel_size': 3,
            'padding': 'same',
            'activation': 'relu'
        },
        'max_pool_layer': {
            'pool_size': 2
        },
        'lstm_layer': {
            'units': 100
        },
        'activation_layer': {
            'activation': 'softmax'
        },
        'optimizer': {
            'module': 'keras.optimizers',
            'name': 'Adam',
            'params': {
                'lr': 1e-3,
                'decay': 0.0
            }
        },
        'compile_params': {
            'loss': 'categorical_crossentropy',
            # 'optimizer': 'adam',
            'metrics': ['accuracy']
        }
    }

    def build_model(self):
        base_model = self.embedding.model
        conv_layer = Conv1D(**self.hyper_parameters['conv_layer'])(base_model.output)
        max_pool_layer = MaxPooling1D(**self.hyper_parameters['max_pool_layer'])(conv_layer)
        lstm_layer = LSTMLayer(**self.hyper_parameters['lstm_layer'])(max_pool_layer)
        dense_layer = Dense(len(self.label2idx),
                            **self.hyper_parameters['activation_layer'])(lstm_layer)
        output_layers = [dense_layer]

        model = Model(base_model.inputs, output_layers)
        optimizer = getattr(eval(self.hyper_parameters['optimizer']['module']),
                            self.hyper_parameters['optimizer']['name'])(
            **self.hyper_parameters['optimizer']['params'])
        model.compile(optimizer=optimizer, **self.hyper_parameters['compile_params'])
        self.model = model
        self.model.summary()


class AVCNNModel(ClassificationModel):
    __architect_name__ = 'AVCNNModel'
    __base_hyper_parameters__ = {
        'spatial_dropout': {
            'rate': 0.25
        },
        'conv_0': {
            'filters': 300,
            'kernel_size': 1,
            'kernel_initializer': 'normal',
            'padding': 'valid',
            'activation': 'relu'
        },
        'conv_1': {
            'filters': 300,
            'kernel_size': 2,
            'kernel_initializer': 'normal',
            'padding': 'valid',
            'activation': 'relu'
        },
        'conv_2': {
            'filters': 300,
            'kernel_size': 3,
            'kernel_initializer': 'normal',
            'padding': 'valid',
            'activation': 'relu'
        },
        'conv_3': {
            'filters': 300,
            'kernel_size': 4,
            'kernel_initializer': 'normal',
            'padding': 'valid',
            'activation': 'relu'
        },
        # ---
        'attn_0': {},
        'avg_0': {},
        'maxpool_0': {},
        # ---
        'maxpool_1': {},
        'attn_1': {},
        'avg_1': {},
        # ---
        'maxpool_2': {},
        'attn_2': {},
        'avg_2': {},
        # ---
        'maxpool_3': {},
        'attn_3': {},
        'avg_3': {},
        # ---
        'v0_col': {
            # 'mode': 'concat',
            'axis': 1
        },
        'v1_col': {
            # 'mode': 'concat',
            'axis': 1
        },
        'v2_col': {
            # 'mode': 'concat',
            'axis': 1
        },
        'merged_tensor': {
            # 'mode': 'concat',
            'axis': 1
        },
        'dropout': {
            'rate': 0.7
        },
        'dense': {
            'units': 144,
            'activation': 'relu'
        },
        'activation_layer': {
            'activation': 'softmax'
        },
        'optimizer': {
            'module': 'keras.optimizers',
            'name': 'Adam',
            'params': {
                'lr': 1e-3,
                'decay': 1e-7
            }
        },
        'compile_params': {
            'loss': 'categorical_crossentropy',
            # 'optimizer': 'adam',
            'metrics': ['accuracy']
        }
    }

    def build_model(self):
        base_model = self.embedding.model
        embedded_seq = SpatialDropout1D(**self.hyper_parameters['spatial_dropout'])(base_model.output)
        conv_0 = Conv1D(**self.hyper_parameters['conv_0'])(embedded_seq)
        conv_1 = Conv1D(**self.hyper_parameters['conv_1'])(embedded_seq)
        conv_2 = Conv1D(**self.hyper_parameters['conv_2'])(embedded_seq)
        conv_3 = Conv1D(**self.hyper_parameters['conv_3'])(embedded_seq)

        maxpool_0 = GlobalMaxPooling1D()(conv_0)
        attn_0 = AttentionWeightedAverage()(conv_0)
        avg_0 = GlobalAveragePooling1D()(conv_0)

        maxpool_1 = GlobalMaxPooling1D()(conv_1)
        attn_1 = AttentionWeightedAverage()(conv_1)
        avg_1 = GlobalAveragePooling1D()(conv_1)

        maxpool_2 = GlobalMaxPooling1D()(conv_2)
        attn_2 = AttentionWeightedAverage()(conv_2)
        avg_2 = GlobalAveragePooling1D()(conv_2)

        maxpool_3 = GlobalMaxPooling1D()(conv_3)
        attn_3 = AttentionWeightedAverage()(conv_3)
        avg_3 = GlobalAveragePooling1D()(conv_3)

        v0_col = concatenate([maxpool_0, maxpool_1, maxpool_2, maxpool_3],
                             **self.hyper_parameters['v0_col'])
        v1_col = concatenate([attn_0, attn_1, attn_2, attn_3],
                             **self.hyper_parameters['v1_col'])
        v2_col = concatenate([avg_1, avg_2, avg_0, avg_3],
                             **self.hyper_parameters['v2_col'])
        merged_tensor = concatenate([v0_col, v1_col, v2_col],
                                    **self.hyper_parameters['merged_tensor'])
        output = Dropout(**self.hyper_parameters['dropout'])(merged_tensor)
        output = Dense(**self.hyper_parameters['dense'])(output)
        output = Dense(len(self.label2idx),
                       **self.hyper_parameters['activation_layer'])(output)

        model = Model(base_model.inputs, output)
        optimizer = getattr(eval(self.hyper_parameters['optimizer']['module']),
                            self.hyper_parameters['optimizer']['name'])(
            **self.hyper_parameters['optimizer']['params'])
        model.compile(optimizer=optimizer, **self.hyper_parameters['compile_params'])
        self.model = model
        self.model.summary()


class KMaxCNNModel(ClassificationModel):
    __architect_name__ = 'KMaxCNNModel'
    __base_hyper_parameters__ = {
        'spatial_dropout': {
            'rate': 0.2
        },
        'conv_0': {
            'filters': 180,
            'kernel_size': 1,
            'kernel_initializer': 'normal',
            'padding': 'valid',
            'activation': 'relu'
        },
        'conv_1': {
            'filters': 180,
            'kernel_size': 2,
            'kernel_initializer': 'normal',
            'padding': 'valid',
            'activation': 'relu'
        },
        'conv_2': {
            'filters': 180,
            'kernel_size': 3,
            'kernel_initializer': 'normal',
            'padding': 'valid',
            'activation': 'relu'
        },
        'conv_3': {
            'filters': 180,
            'kernel_size': 4,
            'kernel_initializer': 'normal',
            'padding': 'valid',
            'activation': 'relu'
        },
        'maxpool_0': {
            'k': 3
        },
        'maxpool_1': {
            'k': 3
        },
        'maxpool_2': {
            'k': 3
        },
        'maxpool_3': {
            'k': 3
        },
        'merged_tensor': {
            # 'mode': 'concat',
            'axis': 1
        },
        'dropout': {
            'rate': 0.6
        },
        'dense': {
            'units': 144,
            'activation': 'relu'
        },
        'activation_layer': {
            'activation': 'softmax'
        },
        'optimizer': {
            'module': 'keras.optimizers',
            'name': 'Adam',
            'params': {
                'lr': 1e-3,
                'decay': 1e-7
            }
        },
        'compile_params': {
            'loss': 'categorical_crossentropy',
            # 'optimizer': 'adam',
            'metrics': ['accuracy']
        }
    }

    def build_model(self):
        base_model = self.embedding.model
        embedded_seq = SpatialDropout1D(**self.hyper_parameters['spatial_dropout'])(base_model.output)
        conv_0 = Conv1D(**self.hyper_parameters['conv_0'])(embedded_seq)
        conv_1 = Conv1D(**self.hyper_parameters['conv_1'])(embedded_seq)
        conv_2 = Conv1D(**self.hyper_parameters['conv_2'])(embedded_seq)
        conv_3 = Conv1D(**self.hyper_parameters['conv_3'])(embedded_seq)

        maxpool_0 = KMaxPooling(**self.hyper_parameters['maxpool_0'])(conv_0)
        # maxpool_0f = Reshape((-1,))(maxpool_0)
        maxpool_0f = Flatten()(maxpool_0)
        maxpool_1 = KMaxPooling(**self.hyper_parameters['maxpool_1'])(conv_1)
        # maxpool_1f = Reshape((-1,))(maxpool_1)
        maxpool_1f = Flatten()(maxpool_1)
        maxpool_2 = KMaxPooling(**self.hyper_parameters['maxpool_2'])(conv_2)
        # maxpool_2f = Reshape((-1,))(maxpool_2)
        maxpool_2f = Flatten()(maxpool_2)
        maxpool_3 = KMaxPooling(**self.hyper_parameters['maxpool_3'])(conv_3)
        # maxpool_3f = Reshape((-1,))(maxpool_3)
        maxpool_3f = Flatten()(maxpool_3)
        # maxpool_0 = GlobalMaxPooling1D()(conv_0)
        # maxpool_1 = GlobalMaxPooling1D()(conv_1)
        # maxpool_2 = GlobalMaxPooling1D()(conv_2)
        # maxpool_3 = GlobalMaxPooling1D()(conv_3)

        # merged_tensor = concatenate([maxpool_0, maxpool_1, maxpool_2, maxpool_3],
        #                            **self.hyper_parameters['merged_tensor'])
        merged_tensor = concatenate([maxpool_0f, maxpool_1f, maxpool_2f, maxpool_3f],
                                    **self.hyper_parameters['merged_tensor'])
        # flatten = Reshape((-1,))(merged_tensor)
        # output = Dropout(**self.hyper_parameters['dropout'])(flatten)
        output = Dropout(**self.hyper_parameters['dropout'])(merged_tensor)
        output = Dense(**self.hyper_parameters['dense'])(output)
        output = Dense(len(self.label2idx),
                       **self.hyper_parameters['activation_layer'])(output)

        model = Model(base_model.inputs, output)
        optimizer = getattr(eval(self.hyper_parameters['optimizer']['module']),
                            self.hyper_parameters['optimizer']['name'])(
            **self.hyper_parameters['optimizer']['params'])
        model.compile(optimizer=optimizer, **self.hyper_parameters['compile_params'])
        self.model = model
        self.model.summary()


class RCNNModel(ClassificationModel):
    __architect_name__ = 'RCNNModel'
    __base_hyper_parameters__ = {
        'spatial_dropout': {
            'rate': 0.2
        },
        'rnn_0': {
            'units': 64,
            'return_sequences': True
        },
        'conv_0': {
            'filters': 128,
            'kernel_size': 2,
            'kernel_initializer': 'normal',
            'padding': 'valid',
            'activation': 'relu',
            'strides': 1
        },
        'maxpool': {},
        'attn': {},
        'average': {},
        'concat': {
            'axis': 1
        },
        'dropout': {
            'rate': 0.5
        },
        'dense': {
            'units': 120,
            'activation': 'relu'
        },
        'activation_layer': {
            'activation': 'softmax'
        },
        'optimizer': {
            'module': 'keras.optimizers',
            'name': 'Adam',
            'params': {
                'lr': 1e-3,
                'clipvalue': 5,
                'decay': 1e-5
            }
        },
        'compile_params': {
            'loss': 'categorical_crossentropy',
            # 'optimizer': 'adam',
            'metrics': ['accuracy']
        }
    }

    def build_model(self):
        base_model = self.embedding.model
        embedded_seq = SpatialDropout1D(**self.hyper_parameters['spatial_dropout'])(base_model.output)
        rnn_0 = Bidirectional(GRULayer(**self.hyper_parameters['rnn_0']))(embedded_seq)
        conv_0 = Conv1D(**self.hyper_parameters['conv_0'])(rnn_0)
        maxpool = GlobalMaxPooling1D()(conv_0)
        attn = AttentionWeightedAverage()(conv_0)
        average = GlobalAveragePooling1D()(conv_0)

        concatenated = concatenate([maxpool, attn, average],
                                   **self.hyper_parameters['concat'])
        output = Dropout(**self.hyper_parameters['dropout'])(concatenated)
        output = Dense(**self.hyper_parameters['dense'])(output)
        output = Dense(len(self.label2idx),
                       **self.hyper_parameters['activation_layer'])(output)

        model = Model(base_model.inputs, output)
        optimizer = getattr(eval(self.hyper_parameters['optimizer']['module']),
                            self.hyper_parameters['optimizer']['name'])(
            **self.hyper_parameters['optimizer']['params'])
        model.compile(optimizer=optimizer, **self.hyper_parameters['compile_params'])
        self.model = model
        self.model.summary()


class AVRNNModel(ClassificationModel):
    __architect_name__ = 'AVRNNModel'
    __base_hyper_parameters__ = {
        'spatial_dropout': {
            'rate': 0.25
        },
        'rnn_0': {
            'units': 60,
            'return_sequences': True
        },
        'rnn_1': {
            'units': 60,
            'return_sequences': True
        },
        'concat_rnn': {
            'axis': 2
        },
        'last': {},
        'maxpool': {},
        'attn': {},
        'average': {},
        'all_views': {
            'axis': 1
        },
        'dropout': {
            'rate': 0.5
        },
        'dense': {
            'units': 144,
            'activation': 'relu'
        },
        'activation_layer': {
            'activation': 'softmax'
        },
        'optimizer': {
            'module': 'keras.optimizers',
            'name': 'Adam',
            'params': {
                'lr': 1e-3,
                'clipvalue': 5,
                'decay': 1e-6
            }
        },
        'compile_params': {
            'loss': 'categorical_crossentropy',
            # 'optimizer': 'adam',
            'metrics': ['accuracy']
        }
    }

    def build_model(self):
        base_model = self.embedding.model
        embedded_seq = SpatialDropout1D(**self.hyper_parameters['spatial_dropout'])(base_model.output)
        rnn_0 = Bidirectional(GRULayer(**self.hyper_parameters['rnn_0']))(embedded_seq)
        rnn_1 = Bidirectional(GRULayer(**self.hyper_parameters['rnn_1']))(rnn_0)
        concat_rnn = concatenate([rnn_0, rnn_1],
                                 **self.hyper_parameters['concat_rnn'])

        last = Lambda(lambda t: t[:, -1], name='last')(concat_rnn)
        maxpool = GlobalMaxPooling1D()(concat_rnn)
        attn = AttentionWeightedAverage()(concat_rnn)
        average = GlobalAveragePooling1D()(concat_rnn)

        all_views = concatenate([last, maxpool, attn, average],
                                **self.hyper_parameters['all_views'])
        output = Dropout(**self.hyper_parameters['dropout'])(all_views)
        output = Dense(**self.hyper_parameters['dense'])(output)
        output = Dense(len(self.label2idx),
                       **self.hyper_parameters['activation_layer'])(output)

        model = Model(base_model.inputs, output)
        optimizer = getattr(eval(self.hyper_parameters['optimizer']['module']),
                            self.hyper_parameters['optimizer']['name'])(
            **self.hyper_parameters['optimizer']['params'])
        model.compile(optimizer=optimizer, **self.hyper_parameters['compile_params'])
        self.model = model
        self.model.summary()


class DropoutBGRUModel(ClassificationModel):
    __architect_name__ = 'DropoutBGRUModel'
    __base_hyper_parameters__ = {
        'spatial_dropout': {
            'rate': 0.15
        },
        'rnn_0': {
            'units': 64,
            'return_sequences': True
        },
        'dropout_rnn': {
            'rate': 0.35
        },
        'rnn_1': {
            'units': 64,
            'return_sequences': True
        },
        'last': {},
        'maxpool': {},
        'average': {},
        'all_views': {
            'axis': 1
        },
        'dropout': {
            'rate': 0.5
        },
        'dense': {
            'units': 72,
            'activation': 'relu'
        },
        'activation_layer': {
            'activation': 'softmax'
        },
        'optimizer': {
            'module': 'keras.optimizers',
            'name': 'Adam',
            'params': {
                'lr': 1e-3,
                'decay': 0.0
            }
        },
        'compile_params': {
            'loss': 'categorical_crossentropy',
            # 'optimizer': 'adam',
            'metrics': ['accuracy']
        }
    }

    def build_model(self):
        base_model = self.embedding.model
        embedded_seq = SpatialDropout1D(**self.hyper_parameters['spatial_dropout'])(base_model.output)
        rnn_0 = Bidirectional(GRULayer(**self.hyper_parameters['rnn_0']))(embedded_seq)
        dropout_rnn = Dropout(**self.hyper_parameters['dropout_rnn'])(rnn_0)
        rnn_1 = Bidirectional(GRULayer(**self.hyper_parameters['rnn_1']))(dropout_rnn)
        last = Lambda(lambda t: t[:, -1], name='last')(rnn_1)
        maxpool = GlobalMaxPooling1D()(rnn_1)
        # attn = AttentionWeightedAverage()(rnn_1)
        average = GlobalAveragePooling1D()(rnn_1)

        all_views = concatenate([last, maxpool, average],
                                **self.hyper_parameters['all_views'])
        output = Dropout(**self.hyper_parameters['dropout'])(all_views)
        output = Dense(**self.hyper_parameters['dense'])(output)
        output = Dense(len(self.label2idx),
                       **self.hyper_parameters['activation_layer'])(output)

        model = Model(base_model.inputs, output)
        optimizer = getattr(eval(self.hyper_parameters['optimizer']['module']),
                            self.hyper_parameters['optimizer']['name'])(
            **self.hyper_parameters['optimizer']['params'])
        model.compile(optimizer=optimizer, **self.hyper_parameters['compile_params'])
        self.model = model
        self.model.summary()


class DropoutAVRNNModel(ClassificationModel):
    __architect_name__ = 'DropoutAVRNNModel'
    __base_hyper_parameters__ = {
        'spatial_dropout': {
            'rate': 0.25
        },
        'rnn_0': {
            'units': 56,
            'return_sequences': True
        },
        'rnn_dropout': {
            'rate': 0.3
        },
        'rnn_1': {
            'units': 56,
            'return_sequences': True
        },
        'last': {},
        'maxpool': {},
        'attn': {},
        'average': {},
        'all_views': {
            'axis': 1
        },
        'dropout_0': {
            'rate': 0.5
        },
        'dense': {
            'units': 128,
            'activation': 'relu'
        },
        'dropout_1': {
            'rate': 0.25
        },
        'activation_layer': {
            'activation': 'softmax'
        },
        'optimizer': {
            'module': 'keras.optimizers',
            'name': 'Adam',
            'params': {
                'lr': 1e-3,
                'clipvalue': 5,
                'decay': 1e-7
            }
        },
        'compile_params': {
            'loss': 'categorical_crossentropy',
            # 'optimizer': 'adam',
            'metrics': ['accuracy']
        }
    }

    def build_model(self):
        base_model = self.embedding.model
        embedded_seq = SpatialDropout1D(**self.hyper_parameters['spatial_dropout'])(base_model.output)
        rnn_0 = Bidirectional(GRULayer(**self.hyper_parameters['rnn_0']))(embedded_seq)
        rnn_dropout = SpatialDropout1D(**self.hyper_parameters['rnn_dropout'])(rnn_0)
        rnn_1 = Bidirectional(GRULayer(**self.hyper_parameters['rnn_1']))(rnn_dropout)

        last = Lambda(lambda t: t[:, -1], name='last')(rnn_1)
        maxpool = GlobalMaxPooling1D()(rnn_1)
        attn = AttentionWeightedAverage()(rnn_1)
        average = GlobalAveragePooling1D()(rnn_1)

        all_views = concatenate([last, maxpool, attn, average],
                                **self.hyper_parameters['all_views'])
        output = Dropout(**self.hyper_parameters['dropout_0'])(all_views)
        output = Dense(**self.hyper_parameters['dense'])(output)
        output = Dropout(**self.hyper_parameters['dropout_1'])(output)
        output = Dense(len(self.label2idx),
                       **self.hyper_parameters['activation_layer'])(output)

        model = Model(base_model.inputs, output)
        optimizer = getattr(eval(self.hyper_parameters['optimizer']['module']),
                            self.hyper_parameters['optimizer']['name'])(
            **self.hyper_parameters['optimizer']['params'])
        model.compile(optimizer=optimizer, **self.hyper_parameters['compile_params'])
        self.model = model
        self.model.summary()


if __name__ == '__main__':
    from kashgari.corpus import TencentDingdangSLUCorpus
    from kashgari.embeddings import WordEmbeddings, BERTEmbedding

    train_x, train_y = TencentDingdangSLUCorpus.get_classification_data()

    w2v = WordEmbeddings('sgns.weibo.bigram',
                         sequence_length=15,
                         limit=5000)
    bert = BERTEmbedding('bert-base-chinese', sequence_length=15)
    t_model = CNNModel(bert)
    t_model.fit(train_x, train_y, epochs=1)



# encoding: utf-8
"""
@author: BrikerMan
@contact: eliyar917@gmail.com
@blog: https://eliyar.biz

@version: 1.0
@license: Apache Licence
@file: __init__.py.py
@time: 2019-01-19 13:42

"""
import kashgari.embeddings
import kashgari.corpus
import kashgari.tasks

from kashgari.tasks import classification
from kashgari.tasks import seq_labeling

from kashgari.macros import config


if __name__ == "__main__":
    print("Hello world")



# encoding: utf-8
"""
@author: BrikerMan
@contact: eliyar917@gmail.com
@blog: https://eliyar.biz

@version: 1.0
@license: Apache Licence
@file: blstm_model
@time: 2019-01-21

"""

import logging

from keras.layers import Bidirectional, LSTM
from keras.layers import Dense, Dropout, TimeDistributed, Activation
from keras.models import Model

from kashgari.tasks.seq_labeling.base_model import SequenceLabelingModel


class BLSTMModel(SequenceLabelingModel):
    __architect_name__ = 'BLSTMModel'
    __base_hyper_parameters__ = {
        'lstm_layer': {
            'units': 256,
            'return_sequences': True
        }, 'dropout_layer': {
            'rate': 0.4
        }
    }

    def build_model(self, loss_f=None, optimizer=None, metrics=None, **kwargs):
        """
        build model function
        :return:
        """
        if not loss_f:
            loss_f = 'categorical_crossentropy'
        if not optimizer:
            optimizer = 'adam'
        if not metrics:
            metrics = ['accuracy']
        embed_model = self.embedding.model

        blstm_layer = Bidirectional(LSTM(**self.hyper_parameters['lstm_layer']))(embed_model.output)
        dropout_layer = Dropout(**self.hyper_parameters['dropout_layer'])(blstm_layer)
        time_distributed_layer = TimeDistributed(Dense(len(self.label2idx)))(dropout_layer)
        activation = Activation('softmax')(time_distributed_layer)

        model = Model(embed_model.inputs, activation)
        model.compile(loss=loss_f,
                      optimizer=optimizer,
                      metrics=metrics)
        self.model = model
        self.model.summary()


if __name__ == '__main__':
    import random
    from keras.callbacks import ModelCheckpoint
    from kashgari.corpus import ChinaPeoplesDailyNerCorpus

    x_train, y_train = ChinaPeoplesDailyNerCorpus.get_sequence_tagging_data()
    x_validate, y_validate = ChinaPeoplesDailyNerCorpus.get_sequence_tagging_data(data_type='validate')
    x_test, y_test = ChinaPeoplesDailyNerCorpus.get_sequence_tagging_data(data_type='test')

    # embedding = WordEmbeddings('sgns.weibo.bigram', sequence_length=100)
    m = BLSTMModel()

    check = ModelCheckpoint('./model.model',
                            monitor='acc',
                            verbose=1,
                            save_best_only=False,
                            save_weights_only=False,
                            mode='auto',
                            period=1)
    m.fit(x_train,
          y_train,
          class_weight=True,
          epochs=1, y_validate=y_validate, x_validate=x_validate, labels_weight=True)

    sample_queries = random.sample(list(range(len(x_train))), 10)
    for i in sample_queries:
        text = x_train[i]
        logging.info('-------- sample {} --------'.format(i))
        logging.info('x: {}'.format(text))
        logging.info('y_true: {}'.format(y_train[i]))
        logging.info('y_pred: {}'.format(m.predict(text)))

    m.evaluate(x_test, y_test)



# encoding: utf-8
"""
@author: BrikerMan
@contact: eliyar917@gmail.com
@blog: https://eliyar.biz

@version: 1.0
@license: Apache Licence
@file: type_hints.py
@time: 2019-01-21 13:55

"""
from typing import Union, List

# ClassificationXType = Union[List[List[str]], List[str]]
# ClassificationYType = List[str]

TextSeqType = List[str]
TokenSeqType = List[int]

TextSeqInputType = Union[List[TextSeqType], TextSeqType]
TokenSeqInputType = Union[List[TokenSeqType], TokenSeqType]


if __name__ == "__main__":
    print("Hello world")



# encoding: utf-8
"""
@author: BrikerMan
@contact: eliyar917@gmail.com
@blog: https://eliyar.biz

@version: 1.0
@license: Apache Licence
@file: __init__.py.py
@time: 2019-01-19 11:49

"""
from .base_model import ClassificationModel
from kashgari.tasks.classification.models import BLSTMModel, CNNLSTMModel, CNNModel
from kashgari.tasks.classification.models import AVCNNModel, KMaxCNNModel, RCNNModel, AVRNNModel
from kashgari.tasks.classification.models import DropoutBGRUModel, DropoutAVRNNModel



# encoding: utf-8
"""
@author: BrikerMan
@contact: eliyar917@gmail.com
@blog: https://eliyar.biz

@version: 1.0
@license: Apache Licence
@file: blstm_crf_model.py
@time: 2019-01-23 17:02

"""
from keras.layers import Dense, Bidirectional
from keras.layers.recurrent import LSTM
from keras.models import Model

from kashgari.utils.crf import CRF, crf_loss, crf_accuracy

from kashgari.tasks.seq_labeling.base_model import SequenceLabelingModel


class BLSTMCRFModel(SequenceLabelingModel):
    __architect_name__ = 'BLSTMCRFModel'
    __base_hyper_parameters__ = {
        'lstm_layer': {
            'units': 256,
            'return_sequences': True
        },
        'dense_layer': {
            'units': 64,
            'activation': 'tanh'
        }
    }

    def build_model(self):
        base_model = self.embedding.model
        blstm_layer = Bidirectional(LSTM(**self.hyper_parameters['lstm_layer']))(base_model.output)
        dense_layer = Dense(128, activation='tanh')(blstm_layer)
        crf = CRF(len(self.label2idx), sparse_target=False)
        crf_layer = crf(dense_layer)
        model = Model(base_model.inputs, crf_layer)
        model.compile(loss=crf_loss,
                      optimizer='adam',
                      metrics=[crf_accuracy])
        self.model = model
        self.model.summary()


if __name__ == "__main__":
    print("Hello world")
    from kashgari.utils.logger import init_logger
    init_logger()
    from kashgari.corpus import ChinaPeoplesDailyNerCorpus

    init_logger()

    x_train, y_train = ChinaPeoplesDailyNerCorpus.get_sequence_tagging_data()
    x_validate, y_validate = ChinaPeoplesDailyNerCorpus.get_sequence_tagging_data(data_type='validate')
    x_test, y_test = ChinaPeoplesDailyNerCorpus.get_sequence_tagging_data(data_type='test')

    tagger = BLSTMCRFModel()
    tagger.fit(x_train, y_train, epochs=2)
    tagger.evaluate(x_validate, y_validate)
    tagger.evaluate(x_test, y_test, debug_info=True)

    model = BLSTMCRFModel.load_model('/Users/brikerman/Downloads/KashgariNER.output/model')
    model.evaluate(x_test, y_test, debug_info=True)


# encoding: utf-8
"""
@author: BrikerMan
@contact: eliyar917@gmail.com
@blog: https://eliyar.biz

@version: 1.0
@license: Apache Licence
@file: base_model.py
@time: 2019-01-27 14:53

"""
import os
import json
import pickle
import pathlib
import traceback
import logging
logger = logging.getLogger(__name__)
import numpy as np
from typing import Dict

import keras
from keras.models import Model
from keras import backend as K

from kashgari.utils import helper
from kashgari.embeddings import CustomEmbedding, BaseEmbedding
from kashgari.utils.crf import CRF, crf_loss, crf_accuracy
from keras_bert.bert import get_custom_objects as get_bert_custom_objects
from kashgari.layers import AttentionWeightedAverage, KMaxPooling, NonMaskingLayer


class BaseModel(object):
    __base_hyper_parameters__ = {}
    __architect_name__ = ''

    @property
    def hyper_parameters(self):
        return self._hyper_parameters_

    def __init__(self, embedding: BaseEmbedding = None, hyper_parameters: Dict = None, **kwargs):
        if embedding is None:
            self.embedding = CustomEmbedding('custom', sequence_length=0, embedding_size=100)
        else:
            self.embedding = embedding
        self.model: Model = None
        self._hyper_parameters_ = self.__base_hyper_parameters__.copy()
        self._label2idx = {}
        self._idx2label = {}
        self.model_info = {}

        self.task = 'classification'

        if hyper_parameters:
            self._hyper_parameters_.update(hyper_parameters)

    def info(self):
        return {
            'architect_name': self.__architect_name__,
            'task': self.task,
            'embedding': self.embedding.info(),
            'hyper_parameters': self.hyper_parameters,
            'model_info': self.model_info
        }

    def save(self, model_path: str):
        pathlib.Path(model_path).mkdir(exist_ok=True, parents=True)

        model_info = self.info()

        with open(os.path.join(model_path, 'labels.json'), 'w', encoding='utf-8') as f:
            f.write(json.dumps(self._label2idx, indent=2, ensure_ascii=False))

        with open(os.path.join(model_path, 'words.json'), 'w', encoding='utf-8') as f:
            f.write(json.dumps(self.embedding.token2idx, indent=2, ensure_ascii=False))

        with open(os.path.join(model_path, 'model.json'), 'w', encoding='utf-8') as f:
            f.write(json.dumps(model_info, indent=2, ensure_ascii=False))

        with open(os.path.join(model_path, 'struct.json'), 'w', encoding='utf-8') as f:
            f.write(self.model.to_json())

        #self.model.save_weights(os.path.join(model_path, 'weights.h5'))
        optimizer_weight_values = None
        try:
            symbolic_weights = getattr(self.model.optimizer, 'weights')
            optimizer_weight_values = K.batch_get_value(symbolic_weights)
        except Exception as e:
            logger.warn('error occur: {}'.format(e))
            traceback.print_tb(e.__traceback__)
            logger.warn('No optimizer weights found.')
        if optimizer_weight_values is not None:
            with open(os.path.join(model_path, 'optimizer.pkl'), 'wb') as f:
                pickle.dump(optimizer_weight_values, f)

        self.model.save(os.path.join(model_path, 'model.model'))
        logger.info('model saved to {}'.format(os.path.abspath(model_path)))

    @staticmethod
    def create_custom_objects(model_info):
        custom_objects = {}
        loss = model_info.get('model_info', {}).get('loss')
        if loss and loss['name'] == 'weighted_categorical_crossentropy':
            loss_f = helper.weighted_categorical_crossentropy(np.array(loss['weights']))
            custom_objects['loss'] = loss_f

        architect_name = model_info.get('architect_name')
        if architect_name and 'CRF' in architect_name:
            custom_objects['CRF'] = CRF
            custom_objects['crf_loss'] = crf_loss
            custom_objects['crf_accuracy'] = crf_accuracy

        embedding = model_info.get('embedding')

        if embedding and embedding['embedding_type'] == 'bert':
            custom_objects['NonMaskingLayer'] = NonMaskingLayer
            custom_objects.update(get_bert_custom_objects())
        custom_objects['AttentionWeightedAverage'] = AttentionWeightedAverage
        custom_objects['KMaxPooling'] = KMaxPooling
        return custom_objects

    @classmethod
    def load_model(cls, model_path: str):
        with open(os.path.join(model_path, 'labels.json'), 'r', encoding='utf-8') as f:
            label2idx = json.load(f)

        with open(os.path.join(model_path, 'words.json'), 'r', encoding='utf-8') as f:
            token2idx = json.load(f)

        with open(os.path.join(model_path, 'model.json'), 'r', encoding='utf-8') as f:
            model_info = json.load(f)
        agent = cls()
        custom_objects = cls.create_custom_objects(model_info)
        agent.model_info = model_info['model_info']
        if custom_objects:
            logger.debug('prepared custom objects: {}'.format(custom_objects))

        try:
            agent.model = keras.models.load_model(os.path.join(model_path, 'model.model'),
                                                  custom_objects=custom_objects)
        except Exception as e:
            logger.warning('Error `{}` occured trying directly model loading. Try to rebuild.'.format(e))
            logger.debug('Load model structure from json.')
            with open(os.path.join(model_path, 'struct.json'), 'r', encoding='utf-8') as f:
                model_struct = f.read()
            agent.model = keras.models.model_from_json(model_struct,
                                                       custom_objects=custom_objects)
            logger.debug('Build optimizer with model info.')
            optimizer_conf = model_info['hyper_parameters'].get('optimizer', None)
            optimizer = 'adam' #default
            if optimizer_conf is not None and isinstance(optimizer_conf, dict):
                module_str = optimizer_conf.get('module', 'None')
                name_str = optimizer_conf.get('name', 'None')
                params = optimizer_conf.get('params', None)
                invalid_set = [None, 'None', '', {}]
                if not any([module_str.strip() in invalid_set,
                            name_str.strip() in invalid_set,
                            params in invalid_set]):
                    try:
                        optimizer = getattr(eval(module_str), name_str)(**params)
                    except:
                        logger.warn('Invalid optimizer configuration in model info. Use `adam` as default.')
            else:
                logger.warn('No optimizer configuration found in model info. Use `adam` as default.')

            default_compile_params = {'loss': 'categorical_crossentropy', 'metrics':['accuracy']}
            compile_params = model_info['hyper_parameters'].get('compile_params', default_compile_params)
            logger.debug('Compile model from scratch.')
            try:
                agent.model.compile(optimizer=optimizer, **compile_params)
            except:
                logger.warn('Failed to compile model. Compile params seems incorrect.')
                logger.warn('Use default options `{}` to compile.'.format(default_compile_params))
                agent.model.compile(optimizer=optimizer, **default_compile_params)
            logger.debug('Load model weights.')
            agent.model.summary()
            agent.model.load_weights(os.path.join(model_path, 'model.model'))
            agent.model._make_train_function()
            optimizer_weight_values = None
            logger.debug('Load optimizer weights.')
            try:
                with open(os.path.join(model_path, 'optimizer.pkl'), 'rb') as f:
                    optimizer_weight_values = pickle.load(f)
            except Exception as e:
                logger.warn('Try to load optimizer weights but no optimizer weights file found.')
            if optimizer_weight_values is not None:
                agent.model.optimizer.set_weights(optimizer_weight_values)
            else:
                logger.warn('Rebuild model but optimizer weights missed. Retrain needed.')
            logger.info('Model rebuild finished.')
        agent.embedding.update(model_info.get('embedding', {}))
        agent.model.summary()
        agent.label2idx = label2idx
        agent.embedding.token2idx = token2idx
        logger.info('loaded model from {}'.format(os.path.abspath(model_path)))
        return agent


if __name__ == "__main__":
    print("Hello world")


# encoding: utf-8
"""
@author: BrikerMan
@contact: eliyar917@gmail.com
@blog: https://eliyar.biz

@version: 1.0
@license: Apache Licence
@file: corpus
@time: 2019-01-20

"""
import logging
import os
import re
from typing import Tuple, List

import pandas as pd

from kashgari.utils import helper

DATA_TRAIN = 'train'
DATA_VALIDATE = 'validate'
DATA_TEST = 'test'


class Corpus(object):
    __corpus_name__ = ''
    __zip_file__name = ''

    __desc__ = ''

    @classmethod
    def get_classification_data(cls,
                                is_test: bool = False,
                                shuffle: bool = True,
                                max_count: int = 0) -> Tuple[List[str], List[str]]:
        pass

    # @classmethod
    # def get_info(cls):
    #     raise NotImplementedError()


class TencentDingdangSLUCorpus(Corpus):

    __corpus_name__ = 'corpus/task-slu-tencent.dingdang-v1.1'
    __zip_file__name = 'corpus/task-slu-tencent.dingdang-v1.1.tar.gz'

    __desc__ = """    Download from NLPCC 2018 Task4 dataset
    details: http://tcci.ccf.org.cn/conference/2018/taskdata.php
    The dataset adopted by this task is a sample of the real query log from a commercial
    task-oriented dialog system. The data is all in Chinese. The evaluation includes three
    domains, namely music, navigation and phone call. Within the dataset, an additional
    domain label ‘OTHERS’ is used to annotate the data not covered by the three domains. To
    simplify the task, we keep only the intents and the slots of high-frequency while ignoring
    others although they appear in the original data. The entire data can be seen as a stream
    of user queries ordered by time stamp. The stream is further split into a series of segments
    according to the gaps of time stamps between queries and each segment is denoted as a
    ‘session’. The contexts within a session are taken into consideration when a query within
    the session was annotated. Below are two example sessions with annotations.
    
    sample
    ```
    1 打电话 phone_call.make_a_phone_call 打电话
    1 我想听美观 music.play 我想听<song>美观</song>
    1 我想听什话 music.play 我想听<song>什话||神话</song>
    1 神话 music.play <song>神话</song>
    
    2 播放调频广播 OTHERS 播放调频广播
    2 给我唱一首一晃就老了 music.play 给我唱一首<song>一晃就老了</song>
    ```
    """

    @classmethod
    def get_info(cls):
        folder_path = helper.cached_path(cls.__corpus_name__, cls.__zip_file__name, )
        logging.info("""{} info\n    dataset path: {}\n{}""".format(cls.__corpus_name__,
                                                                    folder_path,
                                                                    cls.__desc__))

    @classmethod
    def get_classification_data(cls,
                                data_type: str = DATA_TRAIN,
                                shuffle: bool = True,
                                cutter: str = 'char',
                                max_count: int = 0) -> Tuple[List[List[str]], List[str]]:
        """

        :param data_type: {train, validate, test}
        :param shuffle: shuffle or not
        :param cutter:
        :param max_count:
        :return:
        """
        folder_path = helper.cached_path(cls.__corpus_name__,
                                         cls.__zip_file__name)
        if data_type not in [DATA_TRAIN, DATA_VALIDATE, DATA_TEST]:
            raise ValueError('data_type error, please use one onf the {}'.format([DATA_TRAIN,
                                                                                  DATA_VALIDATE,
                                                                                  DATA_TEST]))
        if cutter not in ['char', 'jieba', 'none']:
            raise ValueError('data_type error, please use one onf the {}'.format([DATA_TRAIN,
                                                                                  DATA_VALIDATE,
                                                                                  DATA_TEST]))
        file_path = os.path.join(folder_path, '{}.csv'.format(data_type))
        df = pd.read_csv(file_path)
        x_data = df['text'].values
        y_data = df['domain'].values
        if shuffle:
            x_data, y_data = helper.unison_shuffled_copies(x_data, y_data)

        if max_count != 0:
            x_data = x_data[:max_count]
            y_data = y_data[:max_count]

        if cutter == 'jieba':
            try:
                import jieba
            except ModuleNotFoundError:
                raise ModuleNotFoundError("please install jieba, `$ pip install jieba`")
            x_data = [list(jieba.cut(item)) for item in x_data]
        elif 'char':
            x_data = [list(item) for item in x_data]
        return x_data, y_data

    @staticmethod
    def parse_ner_str(text: str) -> Tuple[str, str]:
        pattern = '<(?P<entity>\\w*)>(?P<value>[^<>]*)<\\/\\w*>'
        x_list = []
        tag_list = []
        last_index = 0
        for m in re.finditer(pattern, text):
            x_list += text[last_index:m.start()]
            tag_list += ['O'] * (m.start() - last_index)
            last_index = m.end()
            dic = m.groupdict()
            value = dic['value'].split('||')[0]
            entity = dic['entity']
            x_list += list(value)
            tag_list += ['P-' + entity] + ['I-' + entity] * (len(value) - 1)
        if last_index < len(text):
            x_list += list(text[last_index:])
            tag_list += len(text[last_index:]) * ['O']
        return ' '.join(x_list), ' '.join(tag_list)

    @classmethod
    def get_sequence_tagging_data(cls,
                                  is_test: bool = False,
                                  shuffle: bool = True,
                                  max_count: int = 0) -> Tuple[List[str], List[str]]:
        folder_path = helper.cached_path(cls.__corpus_name__,
                                         cls.__zip_file__name)

        if is_test:
            file_path = os.path.join(folder_path, 'test.csv')
        else:
            file_path = os.path.join(folder_path, 'train.csv')

        df = pd.read_csv(file_path)
        x_data = []
        y_data = []

        for tagging_text in df['tagging']:
            x_item, y_item = cls.parse_ner_str(tagging_text)
            x_data.append(x_item)
            y_data.append(y_item)
        if shuffle:
            x_data, y_data = helper.unison_shuffled_copies(x_data, y_data)
        if max_count != 0:
            x_data = x_data[:max_count]
            y_data = y_data[:max_count]
        return x_data, y_data


class ChinaPeoplesDailyNerCorpus(object):
    __corpus_name__ = 'corpus/china-people-daily-ner-corpus'
    __zip_file__name = 'corpus/china-people-daily-ner-corpus.tar.gz'

    __desc__ = """
    https://github.com/zjy-ucas/ChineseNER/
    """

    @classmethod
    def get_sequence_tagging_data(cls,
                                  data_type: str = DATA_TRAIN,
                                  shuffle: bool = True,
                                  max_count: int = 0) -> Tuple[List[List[str]], List[List[str]]]:
        folder_path = helper.cached_path(cls.__corpus_name__,
                                         cls.__zip_file__name)

        if data_type == DATA_TRAIN:
            file_path = os.path.join(folder_path, 'example.train')
        elif data_type == DATA_TEST:
            file_path = os.path.join(folder_path, 'example.test')
        else:
            file_path = os.path.join(folder_path, 'example.dev')

        data_x, data_y = [], []

        with open(file_path, 'r', encoding='utf-8') as f:
            lines = f.read().splitlines()
            x, y = [], []
            for line in lines:
                rows = line.split(' ')
                if len(rows) == 1:
                    data_x.append(x)
                    data_y.append(y)
                    x = []
                    y = []
                else:
                    x.append(rows[0])
                    y.append(rows[1])
        return data_x, data_y


class CoNLL2003Corpus(Corpus):
    __corpus_name__ = 'corpus/conll2003'
    __zip_file__name = 'corpus/conll2003.tar.gz'

    @classmethod
    def get_sequence_tagging_data(cls,
                                  data_type: str = DATA_TRAIN,
                                  task_name: str = 'ner',
                                  shuffle: bool = True,
                                  max_count: int = 0) -> Tuple[List[List[str]], List[List[str]]]:
        folder_path = helper.cached_path(cls.__corpus_name__,
                                         cls.__zip_file__name)

        if data_type not in [DATA_TRAIN, DATA_VALIDATE, DATA_TEST]:
            raise ValueError('data_type error, please use one onf the {}'.format([DATA_TRAIN,
                                                                                  DATA_VALIDATE,
                                                                                  DATA_TEST]))
        if task_name not in ['ner', 'pos', 'chunking']:
            raise ValueError('data_type error, please use one onf the {}'.format(['ner', 'pos', 'chunking']))
        folder_path = os.path.join(folder_path, task_name)
        if data_type == DATA_TRAIN:
            file_path = os.path.join(folder_path, 'train.txt')
        elif data_type == DATA_TEST:
            file_path = os.path.join(folder_path, 'test.txt')
        else:
            file_path = os.path.join(folder_path, 'valid.txt')
        x_list, y_list = _load_data_and_labels(file_path)
        if shuffle:
            x_list, y_list = helper.unison_shuffled_copies(x_list, y_list)
        if max_count:
            x_list = x_list[:max_count]
            y_list = y_list[:max_count]
        return x_list, y_list

    __desc__ = """
        http://ir.hit.edu.cn/smp2017ecdt-data
        """


class SMP2017ECDTClassificationCorpus(Corpus):
    __corpus_name__ = 'corpus/smp2017ecdt-data-task1'
    __zip_file__name = 'corpus/smp2017ecdt-data-task1.tar.gz'

    __desc__ = """
    http://ir.hit.edu.cn/smp2017ecdt-data
    """

    @classmethod
    def get_classification_data(cls,
                                data_type: str = DATA_TRAIN,
                                shuffle: bool = True,
                                cutter: str = 'char',
                                max_count: int = 0) -> Tuple[List[List[str]], List[str]]:
        """

        :param data_type: {train, validate, test}
        :param shuffle: shuffle or not
        :param cutter:
        :param max_count:
        :return:
        """
        folder_path = helper.cached_path(cls.__corpus_name__,
                                         cls.__zip_file__name)
        if data_type not in [DATA_TRAIN, DATA_VALIDATE, DATA_TEST]:
            raise ValueError('data_type error, please use one onf the {}'.format([DATA_TRAIN,
                                                                                  DATA_VALIDATE,
                                                                                  DATA_TEST]))
        if cutter not in ['char', 'jieba', 'none']:
            raise ValueError('data_type error, please use one onf the {}'.format([DATA_TRAIN,
                                                                                  DATA_VALIDATE,
                                                                                  DATA_TEST]))

        file_path = os.path.join(folder_path, '{}.csv'.format(data_type))
        df = pd.read_csv(file_path)
        x_data = df['text'].values
        y_data = df['domain'].values
        if shuffle:
            x_data, y_data = helper.unison_shuffled_copies(x_data, y_data)

        if max_count != 0:
            x_data = x_data[:max_count]
            y_data = y_data[:max_count]

        if cutter == 'jieba':
            try:
                import jieba
            except ModuleNotFoundError:
                raise ModuleNotFoundError("please install jieba, `$ pip install jieba`")
            x_data = [list(jieba.cut(item)) for item in x_data]
        elif 'char':
            x_data = [list(item) for item in x_data]
        return x_data, y_data


def _load_data_and_labels(filename, encoding='utf-8'):
    """Loads data and label from a file.
    Args:
        filename (str): path to the file.
        encoding (str): file encoding format.
        The file format is tab-separated values.
        A blank line is required at the end of a sentence.
        For example:
        ```
        EU	B-ORG
        rejects	O
        German	B-MISC
        call	O
        to	O
        boycott	O
        British	B-MISC
        lamb	O
        .	O
        Peter	B-PER
        Blackburn	I-PER
        ...
        ```
    Returns:
        tuple(numpy array, numpy array): data and labels.
    Example:
        >>> filename = 'conll2003/en/ner/train.txt'
        >>> data, labels = load_data_and_labels(filename)
    """
    sents, labels = [], []
    words, tags = [], []
    with open(filename, encoding=encoding) as f:
        for line in f:
            line = line.rstrip()
            if line:
                word, tag = line.split('\t')
                words.append(word)
                tags.append(tag)
            else:
                sents.append(words)
                labels.append(tags)
                words, tags = [], []

    return sents, labels


if __name__ == '__main__':

    # init_logger()
    x, y = CoNLL2003Corpus.get_sequence_tagging_data()
    for i in range(5):
        print('{} -> {}'.format(x[i], y[i]))



#!/usr/bin/env python
# encoding: utf-8
"""
@author: BrikerMan
@contact: eliyar917@gmail.com
@blog: https://eliyar.biz

@version: 1.0
@license: Apache Licence
@file: setup.py
@time: 2019-01-24 16:42

"""
import pathlib

from setuptools import find_packages, setup

from version import __version__

# Package meta-data.
NAME = 'kashgari'
DESCRIPTION = 'Simple and powerful NLP framework, ' \
              'build your state-of-art model in 5 minutes for ' \
              'named entity recognition (NER), part-of-speech ' \
              'tagging (PoS) and text classification tasks.'
URL = 'https://github.com/BrikerMan/Kashgari'
EMAIL = 'eliyar917@gmail.com'
AUTHOR = 'BrikerMan'
LICENSE = 'Apache License 2.0'

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

required = [
    'Keras>=2.2.0',
    'h5py>=2.7.1',
    'keras-bert==0.29.0',
    'scikit-learn>=0.19.1',
    'numpy>=1.14.3',
    'download>=0.3.3',
    'seqeval >=0.0.3',
    'colorlog>=4.0.0',
    'gensim>=3.5.0',
    # 'bz2file>=0.98',
    'sklearn',
    'pandas>=0.23.0'
]

# long_description = ""

setup(
    name=NAME,
    version=__version__,
    description=DESCRIPTION,
    long_description=README,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=EMAIL,
    url=URL,
    packages=find_packages(exclude=('tests',)),
    install_requires=required,
    include_package_data=True,
    license=LICENSE,
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        # 'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy'
    ],
)


if __name__ == "__main__":
    print("Hello world")



# encoding: utf-8
"""
@author: BrikerMan
@contact: eliyar917@gmail.com
@blog: https://eliyar.biz

@version: 1.0
@license: Apache Licence
@file: layers
@time: 2019-02-23

"""
from __future__ import absolute_import, division
import logging

import tensorflow as tf
from keras.layers import Flatten
from keras.layers import GRU, LSTM
from keras.layers import CuDNNGRU, CuDNNLSTM
from keras import initializers
from keras.engine import InputSpec, Layer
from keras import backend as K

from kashgari.macros import config

if config.use_CuDNN_cell:
    GRULayer = CuDNNGRU
    LSTMLayer = CuDNNLSTM
else:
    GRULayer = GRU
    LSTMLayer = LSTM


class AttentionWeightedAverage(Layer):
    '''
    Computes a weighted average of the different channels across timesteps.
    Uses 1 parameter pr. channel to compute the attention value for a single timestep.
    '''

    def __init__(self, return_attention=False, **kwargs):
        self.init = initializers.get('uniform')
        self.supports_masking = True
        self.return_attention = return_attention
        super(AttentionWeightedAverage, self).__init__(**kwargs)

    def build(self, input_shape):
        self.input_spec = [InputSpec(ndim=3)]
        assert len(input_shape) == 3

        self.W = self.add_weight(shape=(input_shape[2], 1),
                                 name='{}_w'.format(self.name),
                                 initializer=self.init)
        self.trainable_weights = [self.W]
        super(AttentionWeightedAverage, self).build(input_shape)

    def call(self, x, mask=None):
        # computes a probability distribution over the timesteps
        # uses 'max trick' for numerical stability
        # reshape is done to avoid issue with Tensorflow
        # and 1-dimensional weights
        logits = K.dot(x, self.W)
        x_shape = K.shape(x)
        logits = K.reshape(logits, (x_shape[0], x_shape[1]))
        ai = K.exp(logits - K.max(logits, axis=-1, keepdims=True))

        # masked timesteps have zero weight
        if mask is not None:
            mask = K.cast(mask, K.floatx())
            ai = ai * mask
        att_weights = ai / (K.sum(ai, axis=1, keepdims=True) + K.epsilon())
        weighted_input = x * K.expand_dims(att_weights)
        result = K.sum(weighted_input, axis=1)
        if self.return_attention:
            return [result, att_weights]
        return result

    def get_output_shape_for(self, input_shape):
        return self.compute_output_shape(input_shape)

    def compute_output_shape(self, input_shape):
        output_len = input_shape[2]
        if self.return_attention:
            return [(input_shape[0], output_len), (input_shape[0], input_shape[1])]
        return (input_shape[0], output_len)

    def compute_mask(self, input, input_mask=None):
        if isinstance(input_mask, list):
            return [None] * len(input_mask)
        else:
            return None


class KMaxPooling(Layer):
    '''
    K-max pooling layer that extracts the k-highest activation from a sequence (2nd dimension).
    TensorFlow backend.

    # Arguments
        k: An int scale,
            indicate k max steps of features to pool.
        sorted: A bool,
            if output is sorted (default) or not.
        data_format: A string,
            one of `channels_last` (default) or `channels_first`.
            The ordering of the dimensions in the inputs.
            `channels_last` corresponds to inputs with shape
            `(batch, steps, features)` while `channels_first`
            corresponds to inputs with shape
            `(batch, features, steps)`.
    # Input shape
        - If `data_format='channels_last'`:
            3D tensor with shape:
            `(batch_size, steps, features)`
        - If `data_format='channels_first'`:
            3D tensor with shape:
            `(batch_size, features, steps)`
    # Output shape
        3D tensor with shape:
        `(batch_size, top-k-steps, features)`
    '''

    def __init__(self, k=1, sorted=True, data_format='channels_last', **kwargs):
        super(KMaxPooling, self).__init__(**kwargs)
        self.input_spec = InputSpec(ndim=3)
        self.k = k
        self.sorted = sorted
        self.data_format = K.normalize_data_format(data_format)

    # def build(self, input_shape):
    #     assert len(input_shape) == 3
    #     super(KMaxPooling, self).build(input_shape)

    def compute_output_shape(self, input_shape):
        if self.data_format == 'channels_first':
            return (input_shape[0], self.k, input_shape[1])
        else:
            return (input_shape[0], self.k, input_shape[2])

    def call(self, inputs):
        if self.data_format == 'channels_last':
            # swap last two dimensions since top_k will be applied along the last dimension
            shifted_input = tf.transpose(inputs, [0, 2, 1])

            # extract top_k, returns two tensors [values, indices]
            top_k = tf.nn.top_k(shifted_input, k=self.k, sorted=self.sorted)[0]
        else:
            top_k = tf.nn.top_k(inputs, k=self.k, sorted=self.sorted)[0]
        # return flattened output
        return tf.transpose(top_k, [0, 2, 1])

    def get_config(self):
        config = {'k': self.k,
                  'sorted': self.sorted,
                  'data_format': self.data_format}
        base_config = super(KMaxPooling, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


class NonMaskingLayer(Layer):
    """
    fix convolutional 1D can't receive masked input, detail: https://github.com/keras-team/keras/issues/4978
    thanks for https://github.com/jacoxu
    """

    def __init__(self, **kwargs):
        self.supports_masking = True
        super(NonMaskingLayer, self).__init__(**kwargs)

    def build(self, input_shape):
        pass

    def compute_mask(self, input, input_mask=None):
        # do not pass the mask to the next layers
        return None

    def call(self, x, mask=None):
        return x

    def get_output_shape_for(self, input_shape):
        return input_shape


if __name__ == '__main__':
    print("hello, world")



# encoding: utf-8
"""
@author: BrikerMan
@contact: eliyar917@gmail.com
@blog: https://eliyar.biz

@version: 1.0
@license: Apache Licence
@file: embedding
@time: 2019-01-20

"""
import json
import logging
import os
from typing import Dict, Any

import keras_bert
import numpy as np
from gensim.models import KeyedVectors
from keras.layers import Input, Embedding
from keras.models import Model
from keras.preprocessing import sequence

import kashgari.macros as k
from kashgari.type_hints import *
from kashgari.utils import helper
from kashgari.layers import NonMaskingLayer

EMBEDDINGS_PATH = os.path.join(k.DATA_PATH, 'embedding')


class BaseEmbedding(object):
    base_dict = {
        k.PAD: 0,
        k.BOS: 1,
        k.EOS: 2,
        k.UNK: 3
    }

    special_tokens = {
        k.PAD: k.PAD,
        k.UNK: k.UNK,
        k.BOS: k.BOS,
        k.EOS: k.EOS
    }

    def __init__(self,
                 name_or_path: str,
                 sequence_length: int = None,
                 embedding_size: int = None,
                 **kwargs):
        """
        init a WordEmbedding
        :param name_or_path: model name as `sgns.weibo.bigram` or model path like '/home/brikerman/w2v.model
        :param sequence_length: length of max sequence, all embedding is shaped as (sequence_length, embedding_size)
        :param embedding_size: embedding vector size, only need to set when using a CustomEmbedding
        :param kwargs: kwargs to pass to the method, func: `BaseEmbedding.build`
        """
        self.embedding_type = 'base'
        self.name = name_or_path
        self.embedding_size = embedding_size
        self._sequence_length = sequence_length
        self.model_path = ''
        self._token2idx: Dict[str, int] = None
        self._idx2token: Dict[int, str] = None
        self._model: Model = None
        self._kwargs = kwargs
        self.build(**kwargs)

    def update(self, info: Dict[str, Any]):
        self.name = info['name']
        self.embedding_type = info['embedding_type']
        self.embedding_size = info['embedding_size']
        self._sequence_length = info['sequence_length']
        self.model_path = info['model_path']
        self._kwargs = info['kwargs']

    def info(self):
        return {
            'embedding_type': self.embedding_type,
            'name': self.name,
            'embedding_size': self.embedding_size,
            'sequence_length': self._sequence_length,
            'model_path': self.model_path,
            'kwargs': self._kwargs
        }

    @property
    def token_count(self):
        return len(self._token2idx)

    @property
    def sequence_length(self):
        return self._sequence_length

    @sequence_length.setter
    def sequence_length(self, val):
        self._sequence_length = val
        self.build(**self._kwargs)

    @property
    def model(self) -> Model:
        return self._model

    @property
    def token2idx(self):
        return self._token2idx

    @property
    def is_bert(self):
        return self.embedding_type == 'bert'

    @token2idx.setter
    def token2idx(self, value):
        self._token2idx = value
        self._idx2token = dict([(value, key) for (key, value) in value.items()])

    @property
    def idx2token(self):
        return self._idx2token

    def build(self, **kwargs):
        raise NotImplementedError()

    def build_token2idx_dict(self, x_data: List[TextSeqType], min_count: int = 5):
        raise NotImplementedError()

    def tokenize(self,
                 sentence: TextSeqInputType,
                 add_bos_eos: bool = True) -> TokenSeqInputType:
        is_list = isinstance(sentence[0], list)

        def tokenize_sentence(text: TextSeqType) -> TokenSeqType:
            tokens = [self.token2idx.get(token, self.token2idx[k.UNK]) for token in text]
            if add_bos_eos:
                tokens = [self.token2idx[k.BOS]] + tokens + [self.token2idx[k.EOS]]
            return tokens

        if is_list:
            return [tokenize_sentence(sen) for sen in sentence]
        else:
            return tokenize_sentence(sentence)

    def embed(self, sentence: TextSeqInputType) -> np.array:
        is_list = isinstance(sentence[0], list)
        tokens = self.tokenize(sentence)

        if is_list:
            embed_input = sequence.pad_sequences(tokens, self.sequence_length, padding='post')
        else:
            embed_input = sequence.pad_sequences([tokens], self.sequence_length, padding='post')

        embed_input = self.prepare_model_input(embed_input)
        print(embed_input)
        embed_pred = self.model.predict(embed_input)
        if is_list:
            return embed_pred
        else:
            return embed_pred[0]

    def prepare_model_input(self, input_x: np.array, **kwargs) -> np.array:
        return input_x


class WordEmbeddings(BaseEmbedding):
    base_dict = {
        k.PAD: 0,
        k.BOS: 1,
        k.EOS: 2,
        k.UNK: 3
    }

    URL_MAP = {
        'sgns.renmin.bigram': 'embedding/word2vec/sgns.renmin.bigram.bz2',
        'sgns.renmin.bigram-char': 'embedding/word2vec/sgns.renmin.bigram-char.bz2',
        'sgns.weibo.bigram': 'embedding/word2vec/sgns.weibo.bigram.bz2',
        'sgns.weibo.bigram-char': 'embedding/word2vec/sgns.weibo.bigram-char.bz2',
    }

    def get_embedding_matrix(self) -> np.array:
        base_matrix = []

        file = os.path.join(k.DATA_PATH, 'w2v_embedding_{}.json'.format(self.embedding_size))
        if os.path.exists(file):
            base_matrix = json.load(open(file, 'r', encoding='utf-8'))
            base_matrix = [np.array(matrix) for matrix in base_matrix]
        else:
            for index, key in enumerate(k.MARKED_KEYS):
                if index != 0:
                    vector = np.random.uniform(-0.5, 0.5, self.embedding_size)
                else:
                    vector = np.zeros(self.embedding_size)
                base_matrix.append(vector)
            with open(file, 'w', encoding='utf-8') as f:
                f.write(json.dumps([list(item) for item in base_matrix]))

        matrix_list = base_matrix + list(self.keyed_vector.vectors)
        return np.array(matrix_list)

    def build(self, **kwargs):
        self.embedding_type = 'word2vec'
        if self.name in WordEmbeddings.URL_MAP:
            url = self.URL_MAP.get(self.name)
            self.name = self.name + '.bz2'
        else:
            url = None

        self.model_path = helper.cached_path(self.name,
                                             url,
                                             sub_folders=['embedding', 'word2vec'])

        self.keyed_vector: KeyedVectors = KeyedVectors.load_word2vec_format(self.model_path, **kwargs)
        self.embedding_size = self.keyed_vector.vector_size

        word2idx = self.base_dict.copy()
        for word in self.keyed_vector.index2entity:
            word2idx[word] = len(word2idx)
        self.token2idx = word2idx

        input_layer = Input(shape=(self.sequence_length,), dtype='int32')
        embedding_matrix = self.get_embedding_matrix()

        current = Embedding(self.token_count,
                            self.embedding_size,
                            input_length=self.sequence_length,
                            weights=[embedding_matrix],
                            trainable=False)(input_layer)
        self._model = Model(input_layer, current)
        logging.debug('------------------------------------------------')
        logging.debug('Loaded gensim word2vec model')
        logging.debug('model        : {}'.format(self.model_path))
        logging.debug('word count   : {}'.format(len(self.keyed_vector.index2entity)))
        logging.debug('Top 50 word  : {}'.format(self.keyed_vector.index2entity[:50]))
        logging.debug('------------------------------------------------')

    def build_token2idx_dict(self, x_data: List[TextSeqType], min_count: int = 5):
        logging.debug("word2vec embedding no need to build token2idx with corpus")


class BERTEmbedding(BaseEmbedding):
    base_dict = {}
    special_tokens = {
        k.PAD: '[PAD]',
        k.UNK: '[UNK]',
        k.BOS: '[CLS]',
        k.EOS: '[SEP]',
    }

    model_key_map = {
        'bert-base-uncased': 'uncased_L-12_H-768_A-12',
        'bert-large-uncased': 'uncased_L-24_H-1024_A-16',
        'bert-base-cased': 'cased_L-12_H-768_A-12',
        'bert-large-cased': 'cased_L-24_H-1024_A-16',
        'bert-base-multilingual-cased': 'multi_cased_L-12_H-768_A-12',
        'bert-base-chinese': 'chinese_L-12_H-768_A-12'
    }

    pre_trained_models = {
        # BERT-Base, Uncased: 12-layer, 768-hidden, 12-heads, 110M parameters
        'uncased_L-12_H-768_A-12': 'https://storage.googleapis.com/bert_models/2018_10_18/'
                                   'uncased_L-12_H-768_A-12.zip',
        # BERT-Large, Uncased
        # 24-layer, 1024-hidden, 16-heads, 340M parameters
        'uncased_L-24_H-1024_A-16': 'https://storage.googleapis.com/bert_models/2018_10_18/'
                                    'uncased_L-24_H-1024_A-16.zip',
        # BERT-Base, Cased
        # 12-layer, 768-hidden, 12-heads , 110M parameters
        'cased_L-12_H-768_A-12': 'https://storage.googleapis.com/bert_models/2018_10_18/'
                                 'cased_L-12_H-768_A-12.zip',
        # BERT-Large, Cased
        # 24-layer, 1024-hidden, 16-heads, 340M parameters
        'cased_L-24_H-1024_A-16': 'https://storage.googleapis.com/bert_models/2018_10_18/'
                                  'cased_L-24_H-1024_A-16.zip',
        # BERT-Base, Multilingual Cased (New, recommended)
        # 104 languages, 12-layer, 768-hidden, 12-heads, 110M parameters
        'multi_cased_L-12_H-768_A-12': 'https://storage.googleapis.com/bert_models/2018_11_23/'
                                       'multi_cased_L-12_H-768_A-12.zip',
        # BERT-Base, Multilingual Uncased (Orig, not recommended)
        # 12-layer, 768-hidden, 12-heads, 110M parameters
        'multilingual_L-12_H-768_A-12': 'https://storage.googleapis.com/bert_models/2018_11_03/'
                                        'multilingual_L-12_H-768_A-12.zip',
        # BERT-Base, Chinese
        # Chinese Simplified and Traditional, 12-layer, 768-hidden, 12-heads, 110M
        'chinese_L-12_H-768_A-12': 'https://storage.googleapis.com/bert_models/2018_11_03/'
                                   'chinese_L-12_H-768_A-12.zip',
    }

    def build(self):
        self.embedding_type = 'bert'
        url = self.pre_trained_models.get(self.model_key_map.get(self.name, self.name))
        self.model_path = helper.cached_path(self.model_key_map.get(self.name, self.name),
                                             url,
                                             ['embedding', 'bert'])

        config_path = os.path.join(self.model_path, 'bert_config.json')
        check_point_path = os.path.join(self.model_path, 'bert_model.ckpt')
        logging.info('loading bert model from {}\n'.format(self.model_path))
        model = keras_bert.load_trained_model_from_checkpoint(config_path,
                                                              check_point_path,
                                                              seq_len=self.sequence_length)
        output_layer = NonMaskingLayer()(model.output)
        self._model = Model(model.inputs, output_layer)

        self.embedding_size = self.model.output_shape[-1]
        dict_path = os.path.join(self.model_path, 'vocab.txt')
        word2idx = {}
        with open(dict_path, 'r', encoding='utf-8') as f:
            words = f.read().splitlines()
        for word in words:
            word2idx[word] = len(word2idx)
        for key, value in self.special_tokens.items():
            word2idx[key] = word2idx[value]

        self.token2idx = word2idx

    def build_token2idx_dict(self, x_data: List[TextSeqType], min_count: int = 5):
        logging.debug("bert embedding no need to build token2idx with corpus")

    def prepare_model_input(self, input_x: np.array, **kwargs) -> np.array:
        input_seg = np.zeros(input_x.shape)
        return [input_x, input_seg]


class CustomEmbedding(BaseEmbedding):
    def __init__(self,
                 name_or_path: str = 'custom-embedding',
                 sequence_length: int = None,
                 embedding_size: int = None,
                 **kwargs):
        """
        :param name_or_path: just a name for custom embedding
        :param sequence_length: length of max sequence, all embedding is shaped as (sequence_length, embedding_size)
        :param embedding_size: embedding vector size, only need to set when using a CustomEmbedding
        :param kwargs: kwargs to pass to the method, func: `BaseEmbedding.build`
        """
        if sequence_length is None or embedding_size is None:
            raise ValueError('Must set sequence_length and sequence_length when using the CustomEmbedding layer')
        super(CustomEmbedding, self).__init__(name_or_path, sequence_length, embedding_size, **kwargs)

    def build(self, **kwargs):
        if self._token2idx is None:
            logging.debug('need to build after build_word2idx')
        else:
            input_x = Input(shape=(self.sequence_length,), dtype='int32')
            current = Embedding(self.token_count,
                                self.embedding_size)(input_x)
            self._model = Model(input_x, current)

    def build_token2idx_dict(self, x_data: List[TextSeqType], min_count: int = 5):
        if self.token2idx is None:
            word_set: Dict[str, int] = {}
            for x_item in x_data:
                for word in x_item:
                    word_set[word] = word_set.get(word, 0) + 1

            word2idx_list = sorted(word_set.items(), key=lambda kv: -kv[1])

            word2idx = self.base_dict.copy()
            for word, count in word2idx_list:
                if count >= min_count:
                    word2idx[word] = len(word2idx)

            self.token2idx = word2idx
        self.build()


if __name__ == '__main__':
    from kashgari.utils.logger import init_logger

    init_logger()
    embedding = WordEmbeddings('sgns.weibo.bigram.bz2', 10)

    sentence = '我 想 去 看 电影www'.split(' ')
    print(embedding.__dict__)

    print(embedding.tokenize(sentence))
    print(embedding.tokenize([sentence]))
    print(embedding.embed([sentence]))



# encoding: utf-8
"""
@author: BrikerMan
@contact: eliyar917@gmail.com
@blog: https://eliyar.biz

@version: 1.0
@license: Apache Licence
@file: __init__.py
@time: 2019-01-21

"""

from .blstm_model import BLSTMModel
from .blstm_crf_model import BLSTMCRFModel
from .cnn_lstm_model import CNNLSTMModel
from .base_model import SequenceLabelingModel


if __name__ == '__main__':
    print("hello, world")



# encoding: utf-8
"""
@author: BrikerMan
@contact: eliyar917@gmail.com
@blog: https://eliyar.biz

@version: 1.0
@license: Apache Licence
@file: cnn_lstm_model.py
@time: 2019-01-24 15:27

"""
import logging

from keras.layers import Dense, Conv1D, TimeDistributed, Activation
from keras.layers.recurrent import LSTM
from keras.models import Model

from kashgari.tasks.seq_labeling.base_model import SequenceLabelingModel


class CNNLSTMModel(SequenceLabelingModel):
    __architect_name__ = 'CNNLSTMModel'
    __base_hyper_parameters__ = {
        'conv_layer': {
            'filters': 32,
            'kernel_size': 3,
            'padding': 'same',
            'activation': 'relu'
        },
        'max_pool_layer': {
            'pool_size': 2
        },
        'lstm_layer': {
            'units': 100,
            'return_sequences': True
        }
    }

    def build_model(self):
        base_model = self.embedding.model
        conv_layer = Conv1D(**self.hyper_parameters['conv_layer'])(base_model.output)
        # max_pool_layer = MaxPooling1D(**self.hyper_parameters['max_pool_layer'])(conv_layer)
        lstm_layer = LSTM(**self.hyper_parameters['lstm_layer'])(conv_layer)
        time_distributed_layer = TimeDistributed(Dense(len(self.label2idx)))(lstm_layer)
        activation = Activation('softmax')(time_distributed_layer)
        output_layers = [activation]

        model = Model(base_model.inputs, output_layers)
        model.compile(loss='categorical_crossentropy',
                      optimizer='adam',
                      metrics=['accuracy'])
        self.model = model
        self.model.summary()


if __name__ == "__main__":
    print("Hello world")
    from kashgari.utils.logger import init_logger
    from kashgari.corpus import ChinaPeoplesDailyNerCorpus

    init_logger()

    x_train, y_train = ChinaPeoplesDailyNerCorpus.get_sequence_tagging_data()
    x_validate, y_validate = ChinaPeoplesDailyNerCorpus.get_sequence_tagging_data(data_type='validate')
    x_test, y_test = ChinaPeoplesDailyNerCorpus.get_sequence_tagging_data(data_type='test')

    classifier = CNNLSTMModel()
    classifier.fit(x_train, y_train, epochs=2)
    classifier.evaluate(x_validate, y_validate)
    classifier.evaluate(x_test, y_train)
